import glob
import os.path
import traceback
import hmac
import time
import requests
import sys
sys.path.append('/home/work_nfs7/xlgeng/code_runner_gxl/gxl_ai_utils')

from gxl_ai_utils.thread.my_thread import GxlDynamicThreadPool

import gxl_ai_utils.AiConstant
from gxl_ai_utils.utils import utils_spider, utils_file

# logging = gxl_ai_utils.AiConstant.AI_logging()
import logging
utils_file.set_logging()

class GxlSpider:
    def __init__(self, output_dir='./output/qingting_fm/'):
        self.access_token = None
        self.qingting_id = None
        self.login_url = "https://user.qtfm.cn/u2/api/v4/user/login"
        self.root_url = 'https://www.qtfm.cn'
        self.start_url = 'https://www.qtfm.cn/categories'
        self.output_dir = output_dir
        utils_file.makedir_sil(self.output_dir)

    def handle_root_category(self):
        """得到每个类别的入口"""
        xpath_href = '/html/body/div[1]/div/div[3]/div/div/div[1]/ul/li/a/@href'
        xpath_catefory_title = '/html/body/div[1]/div/div[3]/div/div/div[1]/ul/li/a/text()'
        response = utils_spider.send_request(self.start_url)
        res_href = utils_spider.handle_xpath(response.text, xpath_href)
        res_title = utils_spider.handle_xpath(response.text, xpath_catefory_title)
        res_dict = {}
        for href, title in zip(res_href, res_title):
            res_dict[title] = utils_file.join_path(self.root_url, href)
        utils_file.write_dict_to_json(res_dict, self.output_dir + 'root_category.json')

    def handle_category_index_fenye(self):
        category_dict = utils_file.load_dict_from_json(self.output_dir + 'root_category.json')
        for title, href in category_dict.items():
            utils_file.logging_print(f'开始处理{title}类别')
            if title == '播客' or title == '小说':
                """播客的音频都是'该音频不存在'的提示音。 小说页面的结构不一样"""
                continue
            href = href[:-1]
            album_path_list = []
            save_path = utils_file.join_path(self.output_dir, 'album_info', f'{title}.list')
            utils_file.makedir_for_file_or_dir(save_path)
            for i in range(1, 76):
                utils_file.logging_print(f'开始处理{title}类别的第{i}页')
                try:
                    temp_href = utils_file.join_path(href, str(i))
                    utils_file.logging_print(f'开始处理的页的地址:{temp_href}')
                    xpath_href = "/html/body/div[1]/div/div[3]/div/div/div[2]/div[2]/div/a[1]/@href"
                    response = utils_spider.send_request(temp_href)
                    res_href = utils_spider.handle_xpath(response.text, xpath_href)
                    for i, href_i in enumerate(res_href):
                        href_i = utils_file.join_path(self.root_url, href_i)
                        utils_file.logging_print(href_i)
                        album_path_list.append(href_i)
                except Exception as e:
                    utils_file.logging_print(f'分页{i}的handle_category_index函数执行报错, 爆错内容为{e}')
                    traceback.print_exc()
            utils_file.write_list_to_file(album_path_list, save_path)
            # break

    def handle_ibum_index_fenye(self, ):
        """把一个专辑的音频拿下"""
        album_info_dir = utils_file.join_path(self.output_dir, 'album_info')
        category_s_album_list = glob.glob(utils_file.join_path(album_info_dir, '*.list'))
        utils_file.print_list(category_s_album_list)
        for album_list_path in category_s_album_list:
            category_name = utils_file.get_file_pure_name_from_path(album_list_path)
            album_list = utils_file.load_list_file_clean(album_list_path)
            album_info_category_dir = utils_file.join_path(self.output_dir, 'wav_info', category_name)
            for album_href in album_list:
                album_id = album_href.split('/')[-1]
                album_save_path = utils_file.join_path(album_info_category_dir, f'{album_id}.list')
                res_list = []
                utils_file.logging_print(f'开始处理专辑, url:{album_href}')
                try:
                    xpath_total_page = '/html/body/div[1]/div/div[3]/div[2]/div[1]/div[2]/div/div[3]/div/ul/li/a/text()'
                    res = utils_spider.handle_xpath(utils_spider.send_request(album_href).text, xpath_total_page)
                    total_page = int(res[-2])
                    utils_file.logging_print(f'总页数为{total_page}')
                    for i in range(1, total_page + 1):
                        utils_file.logging_print(f'开始处理第{i}页')
                        try:
                            temp_href = utils_file.join_path(album_href, f'{i}')

                            xpath_wav_list = '/html/body/div[1]/div/div[3]/div[2]/div[1]/div[2]/div/ul/li/span[1]/a/@href'
                            title_wav_list = '/html/body/div[1]/div/div[3]/div[2]/div[1]/div[2]/div/ul/li/span[1]/a/p/text()'
                            res_wav_list = utils_spider.hande_href(temp_href, xpath_wav_list)
                            title_wav_list = utils_spider.hande_href(temp_href, title_wav_list)
                            for wav, title in zip(res_wav_list, title_wav_list):
                                wav_href = utils_file.join_path(self.root_url, wav)
                                res_list.append(wav_href)
                                utils_file.logging_print(wav_href)
                        except Exception as e:
                            utils_file.logging_print(f'handle_ibum_index_fenye函数执行报错, 爆错内容为{e}')
                            traceback.print_exc()
                    utils_file.write_list_to_file(res_list, album_save_path)
                except Exception as e:
                    utils_file.logging_print(f'handle_ibum_index_fenye函数执行报错, 爆错内容为{e}')
                    traceback.print_exc()
                # break

    def get_wav_info(self):
        album_info_dir = utils_file.join_path(self.output_dir, 'wav_info')
        album_path_list = glob.glob(utils_file.join_path(album_info_dir, '**/*.list'))
        utils_file.print_list(album_path_list)
        for album_path in album_path_list:
            self.login()
            album_path = utils_file.normal_path(album_path)
            utils_file.logging_print(f'\n')
            utils_file.logging_print(f'\n')
            utils_file.logging_print(f'\n')
            utils_file.logging_print(f'new album_path, album_path: {album_path}')
            category_name = album_path.split('/')[-2]
            utils_file.logging_print(category_name)
            wav_path_list = utils_file.load_list_file_clean(album_path)
            index = 0
            for wav_path in wav_path_list:
                utils_file.logging_print(wav_path)
                wav_id = wav_path.split('/')[-1]
                album_id = wav_path.split('/')[-3]
                self.get_wav_whole_url(album_id, wav_id, category_name)
                # index += 1
                # if index % 10 == 0:
                #     self.login()

    def login(self, user_id='13937825104', password='Gxl2236035.'):
        data = {
            'account_type': '5',
            'device_id': 'web',
            'user_id': user_id,
            'password': password
        }
        response = requests.post(self.login_url, data=data, headers=utils_spider.COMMON_HEADER)
        if response.status_code == 200:
            temp = response.json()
            errorno = temp['errorno']
            errormsg = temp['errormsg']
            if errorno == 0:
                utils_file.logging_print('login successful!')
                data = temp['data']
                self.qingting_id = data['qingting_id']
                self.access_token = data['access_token']

            else:
                utils_file.logging_print('Login failed')
                print(errormsg)
        self.print_token()

    def print_token(self):
        utils_file.logging_print(self.access_token)
        utils_file.logging_print(self.qingting_id)

    def get_wav_whole_url(self, album_id='474184', wav_id='25643023', category_name='音乐'):
        base_url = "https://audio.qingting.fm"
        bookid = album_id
        id = wav_id
        wav_path = utils_file.join_path(self.output_dir, 'wav', category_name, bookid,
                                        f'{wav_id}_{bookid}.mp3')
        utils_file.logging_print(f'开始得到如下音频的url,wav_path: {wav_path}')
        if os.path.exists(wav_path):
            utils_file.logging_print(f'{wav_id}_{bookid}已存在,不再下载')
            return
        access_token = ""
        qingting_id = ""
        timestamp = str(round(time.time() * 1000))
        data = f"/audiostream/redirect/{bookid}/{id}?access_token={access_token}&device_id=MOBILESITE&qingting_id={qingting_id}&t={timestamp}"
        message = data.encode('utf-8')
        key = "fpMn12&38f_2e".encode('utf-8')
        sign = hmac.new(key, message, digestmod='MD5').hexdigest()
        whole_url = base_url + data + "&sign=" + sign
        utils_file.logging_print(f'得到url, 如下：{whole_url}')
        try:
            utils_file.download_file(whole_url, utils_file.join_path(self.output_dir, 'wav', category_name, bookid),
                                     f'{wav_id}_{bookid}', 'mp3')
        except Exception as e:
            utils_file.logging_print(f'get_wav_whole_url函数执行报错, 爆错内容为{e}')
            # traceback.print_exc()


if __name__ == '__main__':
    gxl_spider = GxlSpider('/home/node36_data/qingting_fm/')
    # gxl_spider = GxlSpider('./output/qingting_fm/')
    # gxl_spider.handle_root_category()
    # gxl_spider.handle_category_index_fenye()
    # gxl_spider.handle_ibum_index_fenye()
    gxl_spider.get_wav_info()
    # gxl_spider.login()
    # gxl_spider.print_token()
    # gxl_spider.get_wav_whole_url()



from typing import List, Optional, Tuple

import torch
from torch import nn
from ..utils.common import get_activation, get_rnn


def ApplyPadding(input, padding, pad_value) -> torch.Tensor:
    """
    Args:
        input:   [bs, max_time_step, dim]
        padding: [bs, max_time_step]
    """
    return padding * pad_value + input * (1 - padding)


class PredictorBase(torch.nn.Module):

    # NOTE(Mddct): We can use ABC abstract here, but
    # keep this class simple enough for now
    def __init__(self) -> None:
        super().__init__()

    def init_state(self,
                   batch_size: int,
                   device: torch.device,
                   method: str = "zero") -> List[torch.Tensor]:
        _, _, _ = batch_size, method, device
        raise NotImplementedError("this is a base precictor")

    def batch_to_cache(self,
                       cache: List[torch.Tensor]) -> List[List[torch.Tensor]]:
        _ = cache
        raise NotImplementedError("this is a base precictor")

    def cache_to_batch(self,
                       cache: List[List[torch.Tensor]]) -> List[torch.Tensor]:
        _ = cache
        raise NotImplementedError("this is a base precictor")

    def output_size(self):
        raise NotImplementedError("this is a base precictor")

    def forward(
        self,
        input: torch.Tensor,
        cache: Optional[List[torch.Tensor]] = None,
    ):
        _, _, = input, cache
        raise NotImplementedError("this is a base precictor")

    def forward_step(
            self, input: torch.Tensor, padding: torch.Tensor,
            cache: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        _, _, _, = input, padding, cache
        raise NotImplementedError("this is a base precictor")


class RNNPredictor(PredictorBase):

    def __init__(self,
                 voca_size: int,
                 embed_size: int,
                 output_size: int,
                 embed_dropout: float,
                 hidden_size: int,
                 num_layers: int,
                 bias: bool = True,
                 rnn_type: str = "lstm",
                 dropout: float = 0.1) -> None:
        super().__init__()
        self.n_layers = num_layers
        self.hidden_size = hidden_size
        self._output_size = output_size
        # disable rnn base out projection
        self.embed = nn.Embedding(voca_size, embed_size)
        self.dropout = nn.Dropout(embed_dropout)
        # NOTE(Mddct): rnn base from torch not support layer norm
        # will add layer norm and prune value in cell and layer
        # ref: https://github.com/Mddct/neural-lm/blob/main/models/gru_cell.py
        self.rnn = get_rnn(rnn_type=rnn_type)(input_size=embed_size,
                                              hidden_size=hidden_size,
                                              num_layers=num_layers,
                                              bias=bias,
                                              batch_first=True,
                                              dropout=dropout)
        self.projection = nn.Linear(hidden_size, output_size)

    def output_size(self):
        return self._output_size

    def forward(
        self,
        input: torch.Tensor,
        cache: Optional[List[torch.Tensor]] = None,
    ) -> torch.Tensor:
        """
        Args:
            input (torch.Tensor): [batch, max_time).
            padding (torch.Tensor): [batch, max_time]
            cache : rnn predictor cache[0] == state_m
                    cache[1] == state_c
        Returns:
            output: [batch, max_time, output_size]
        """

        # NOTE(Mddct): we don't use pack input format
        embed = self.embed(input)  # [batch, max_time, emb_size]
        embed = self.dropout(embed)
        states: Optional[Tuple[torch.Tensor, torch.Tensor]] = None
        if cache is None:
            state = self.init_state(batch_size=input.size(0),
                                    device=input.device)
            states = (state[0], state[1])
        else:
            assert len(cache) == 2
            states = (cache[0], cache[1])
        out, (m, c) = self.rnn(embed, states)
        out = self.projection(out)

        # NOTE(Mddct): Although we don't use staate in transducer
        # training forward, we need make it right for padding value
        # so we create forward_step for infering, forward for training
        _, _ = m, c
        return out

    def batch_to_cache(self,
                       cache: List[torch.Tensor]) -> List[List[torch.Tensor]]:
        """
        Args:
           cache: [state_m, state_c]
               state_ms: [1*n_layers, bs, ...]
               state_cs: [1*n_layers, bs, ...]
        Returns:
           new_cache: [[state_m_1, state_c_1], [state_m_2, state_c_2]...]
        """
        assert len(cache) == 2
        state_ms = cache[0]
        state_cs = cache[1]

        assert state_ms.size(1) == state_cs.size(1)

        new_cache: List[List[torch.Tensor]] = []
        for state_m, state_c in zip(torch.split(state_ms, 1, dim=1),
                                    torch.split(state_cs, 1, dim=1)):
            new_cache.append([state_m, state_c])
        return new_cache

    def cache_to_batch(self,
                       cache: List[List[torch.Tensor]]) -> List[torch.Tensor]:
        """
        Args:
            cache : [[state_m_1, state_c_1], [state_m_1, state_c_1]...]

        Returns:
            new_caceh: [state_ms, state_cs],
                state_ms: [1*n_layers, bs, ...]
                state_cs: [1*n_layers, bs, ...]
        """
        state_ms = torch.cat([states[0] for states in cache], dim=1)
        state_cs = torch.cat([states[1] for states in cache], dim=1)
        return [state_ms, state_cs]

    def init_state(
        self,
        batch_size: int,
        device: torch.device,
        method: str = "zero",
    ) -> List[torch.Tensor]:
        assert batch_size > 0
        # TODO(Mddct): xavier init method
        _ = method
        return [
            torch.zeros(1 * self.n_layers,
                        batch_size,
                        self.hidden_size,
                        device=device),
            torch.zeros(1 * self.n_layers,
                        batch_size,
                        self.hidden_size,
                        device=device)
        ]

    def forward_step(
            self, input: torch.Tensor, padding: torch.Tensor,
            cache: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """
        Args:
            input (torch.Tensor): [batch_size, time_step=1]
            padding (torch.Tensor): [batch_size,1], 1 is padding value
            cache : rnn predictor cache[0] == state_m
                    cache[1] == state_c
        """
        assert len(cache) == 2
        state_m, state_c = cache[0], cache[1]
        embed = self.embed(input)  # [batch, 1, emb_size]
        embed = self.dropout(embed)
        out, (m, c) = self.rnn(embed, (state_m, state_c))

        out = self.projection(out)
        m = ApplyPadding(m, padding.unsqueeze(0), state_m)
        c = ApplyPadding(c, padding.unsqueeze(0), state_c)

        return (out, [m, c])


class EmbeddingPredictor(PredictorBase):
    """Embedding predictor

    Described in:
    https://arxiv.org/pdf/2109.07513.pdf

    embed-> proj -> layer norm -> swish
    """

    def __init__(self,
                 voca_size: int,
                 embed_size: int,
                 embed_dropout: float,
                 n_head: int,
                 history_size: int = 2,
                 activation: str = "swish",
                 bias: bool = False,
                 layer_norm_epsilon: float = 1e-5) -> None:

        super().__init__()
        # multi head
        self.num_heads = n_head
        self.embed_size = embed_size
        self.context_size = history_size + 1
        self.pos_embed = torch.nn.Linear(embed_size * self.context_size,
                                         self.num_heads,
                                         bias=bias)
        self.embed = nn.Embedding(voca_size, self.embed_size)
        self.embed_dropout = nn.Dropout(p=embed_dropout)
        self.ffn = nn.Linear(self.embed_size, self.embed_size)
        self.norm = nn.LayerNorm(self.embed_size, eps=layer_norm_epsilon)
        self.activatoin = get_activation(activation)

    def output_size(self):
        return self.embed_size

    def init_state(self,
                   batch_size: int,
                   device: torch.device,
                   method: str = "zero") -> List[torch.Tensor]:
        assert batch_size > 0
        _ = method
        return [
            torch.zeros(batch_size,
                        self.context_size - 1,
                        self.embed_size,
                        device=device),
        ]

    def batch_to_cache(self,
                       cache: List[torch.Tensor]) -> List[List[torch.Tensor]]:
        """
        Args:
            cache : [history]
                history: [bs, ...]
        Returns:
            new_ache : [[history_1], [history_2], [history_3]...]
        """
        assert len(cache) == 1
        cache_0 = cache[0]
        history: List[List[torch.Tensor]] = []
        for h in torch.split(cache_0, 1, dim=0):
            history.append([h])
        return history

    def cache_to_batch(self,
                       cache: List[List[torch.Tensor]]) -> List[torch.Tensor]:
        """
        Args:
            cache : [[history_1], [history_2], [history3]...]

        Returns:
            new_caceh: [history],
                history: [bs, ...]
        """
        history = torch.cat([h[0] for h in cache], dim=0)
        return [history]

    def forward(self,
                input: torch.Tensor,
                cache: Optional[List[torch.Tensor]] = None):
        """ forward for training
        """
        input = self.embed(input)  # [bs, seq_len, embed]
        input = self.embed_dropout(input)
        if cache is None:
            zeros = self.init_state(input.size(0), device=input.device)[0]
        else:
            assert len(cache) == 1
            zeros = cache[0]

        input = torch.cat((zeros, input),
                          dim=1)  # [bs, context_size-1 + seq_len, embed]

        input = input.unfold(1, self.context_size, 1).permute(
            0, 1, 3, 2)  # [bs, seq_len, context_size, embed]
        # multi head pos: [n_head, embed, context_size]
        multi_head_pos = self.pos_embed.weight.view(self.num_heads,
                                                    self.embed_size,
                                                    self.context_size)

        # broadcast dot attenton
        input_expand = input.unsqueeze(
            2)  # [bs, seq_len, 1, context_size, embed]
        multi_head_pos = multi_head_pos.permute(
            0, 2, 1)  # [num_heads, context_size, embed]

        # [bs, seq_len, num_heads, context_size, embed]
        weight = input_expand * multi_head_pos
        weight = weight.sum(dim=-1, keepdim=False).unsqueeze(
            3)  # [bs, seq_len, num_heads, 1, context_size]
        output = weight.matmul(input_expand).squeeze(
            dim=3)  # [bs, seq_len, num_heads, embed]
        output = output.sum(dim=2)  # [bs, seq_len, embed]
        output = output / (self.num_heads * self.context_size)

        output = self.ffn(output)
        output = self.norm(output)
        output = self.activatoin(output)
        return output

    def forward_step(
        self,
        input: torch.Tensor,
        padding: torch.Tensor,
        cache: List[torch.Tensor],
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """ forward step for inference
        Args:
            input (torch.Tensor): [batch_size, time_step=1]
            padding (torch.Tensor): [batch_size,1], 1 is padding value
            cache: for embedding predictor, cache[0] == history
        """
        assert input.size(1) == 1
        assert len(cache) == 1
        history = cache[0]
        assert history.size(1) == self.context_size - 1
        input = self.embed(input)  # [bs, 1, embed]
        input = self.embed_dropout(input)
        context_input = torch.cat((history, input), dim=1)
        input_expand = context_input.unsqueeze(1).unsqueeze(
            2)  # [bs, 1, 1, context_size, embed]

        # multi head pos: [n_head, embed, context_size]
        multi_head_pos = self.pos_embed.weight.view(self.num_heads,
                                                    self.embed_size,
                                                    self.context_size)

        multi_head_pos = multi_head_pos.permute(
            0, 2, 1)  # [num_heads, context_size, embed]
        # [bs, 1, num_heads, context_size, embed]
        weight = input_expand * multi_head_pos
        weight = weight.sum(dim=-1, keepdim=False).unsqueeze(
            3)  # [bs, 1, num_heads, 1, context_size]
        output = weight.matmul(input_expand).squeeze(
            dim=3)  # [bs, 1, num_heads, embed]
        output = output.sum(dim=2)  # [bs, 1, embed]
        output = output / (self.num_heads * self.context_size)

        output = self.ffn(output)
        output = self.norm(output)
        output = self.activatoin(output)
        new_cache = context_input[:, 1:, :]
        # TODO(Mddct): we need padding new_cache in future
        # new_cache = ApplyPadding(history, padding, new_cache)
        return (output, [new_cache])


class ConvPredictor(PredictorBase):

    def __init__(self,
                 voca_size: int,
                 embed_size: int,
                 embed_dropout: float,
                 history_size: int = 2,
                 activation: str = "relu",
                 bias: bool = False,
                 layer_norm_epsilon: float = 1e-5) -> None:
        super().__init__()

        assert history_size >= 0
        self.embed_size = embed_size
        self.context_size = history_size + 1
        self.embed = nn.Embedding(voca_size, self.embed_size)
        self.embed_dropout = nn.Dropout(p=embed_dropout)
        self.conv = nn.Conv1d(in_channels=embed_size,
                              out_channels=embed_size,
                              kernel_size=self.context_size,
                              padding=0,
                              groups=embed_size,
                              bias=bias)
        self.norm = nn.LayerNorm(embed_size, eps=layer_norm_epsilon)
        self.activatoin = get_activation(activation)

    def output_size(self):
        return self.embed_size

    def init_state(self,
                   batch_size: int,
                   device: torch.device,
                   method: str = "zero") -> List[torch.Tensor]:
        assert batch_size > 0
        assert method == "zero"
        return [
            torch.zeros(batch_size,
                        self.context_size - 1,
                        self.embed_size,
                        device=device)
        ]

    def cache_to_batch(self,
                       cache: List[List[torch.Tensor]]) -> List[torch.Tensor]:
        """
        Args:
            cache : [[history_1], [history_2], [history3]...]

        Returns:
            new_caceh: [history],
                history: [bs, ...]
        """
        history = torch.cat([h[0] for h in cache], dim=0)
        return [history]

    def batch_to_cache(self,
                       cache: List[torch.Tensor]) -> List[List[torch.Tensor]]:
        """
        Args:
            cache : [history]
                history: [bs, ...]
        Returns:
            new_ache : [[history_1], [history_2], [history_3]...]
        """
        assert len(cache) == 1
        cache_0 = cache[0]
        history: List[List[torch.Tensor]] = []
        for h in torch.split(cache_0, 1, dim=0):
            history.append([h])
        return history

    def forward(self,
                input: torch.Tensor,
                cache: Optional[List[torch.Tensor]] = None):
        """ forward for training
        """
        input = self.embed(input)  # [bs, seq_len, embed]
        input = self.embed_dropout(input)
        if cache is None:
            zeros = self.init_state(input.size(0), device=input.device)[0]
        else:
            assert len(cache) == 1
            zeros = cache[0]

        input = torch.cat((zeros, input),
                          dim=1)  # [bs, context_size-1 + seq_len, embed]
        input = input.permute(0, 2, 1)
        out = self.conv(input).permute(0, 2, 1)
        out = self.activatoin(self.norm(out))
        return out

    def forward_step(
            self, input: torch.Tensor, padding: torch.Tensor,
            cache: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """ forward step for inference
        Args:
            input (torch.Tensor): [batch_size, time_step=1]
            padding (torch.Tensor): [batch_size,1], 1 is padding value
            cache: for embedding predictor, cache[0] == history
        """
        assert input.size(1) == 1
        assert len(cache) == 1
        history = cache[0]
        assert history.size(1) == self.context_size - 1
        input = self.embed(input)  # [bs, 1, embed]
        input = self.embed_dropout(input)
        context_input = torch.cat((history, input), dim=1)
        input = context_input.permute(0, 2, 1)
        out = self.conv(input).permute(0, 2, 1)
        out = self.activatoin(self.norm(out))

        new_cache = context_input[:, 1:, :]
        # TODO(Mddct): apply padding in future
        return (out, [new_cache])



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .test import GxlHa


def print_name(obj: "GxlHa"):
    """
    TYPE_CHECKING中引入的内容需要加上“”才能使用
    :param obj:
    :return:
    """
    print(obj.name)



import codecs
import glob
import io
import multiprocessing
import os
import re
import sys
import tarfile
from io import StringIO
from logging import Logger

import types
import torch
import torchaudio
import wave
import platform
from torch.utils.data import DataLoader, Dataset
import torchaudio.compliance.kaldi as kaldi
from tqdm import tqdm
import logging
from ..config.gxl_config import GxlNode

# if platform.system().lower() == 'windows':
#     torchaudio.set_audio_backend("soundfile")
# elif platform.system().lower() == 'linux':
#     torchaudio.set_audio_backend("sox_io")
# else:
#     pass

HAS_SET_LOGGING = False


def set_logging():
    """
    设置日志输出的格式
    :return: 
    """
    global HAS_SET_LOGGING
    HAS_SET_LOGGING = True
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)s %(message)s')


def logging_print(*args):
    global HAS_SET_LOGGING
    if not HAS_SET_LOGGING:
        set_logging()
    string_temp = " ".join([str(arg) for arg in args])
    logging.info(string_temp)
    
    
is_python2 = sys.version_info[0] == 2


def torchaudio_load(wav_path: str, normalize=True, **kwargs):
    """
    
    :param normalize:
    :param wav_path:
    :param kwargs: 
    :return: waveform: torch.tensor: shape(1, sample_num)
    """
    waveform, sr = torchaudio.load(wav_path, normalize=normalize, **kwargs)
    return waveform, sr


def torchaudio_save(save_path: str, waveform, sr, is_normal_waveform: bool = True, **kwargs):
    if is_normal_waveform:
        waveform = waveform * (1 << 15)
    if platform.system().lower() == 'windows':
        torchaudio.save(save_path, waveform, sr, bits_per_sample=16, channels_first=True, format="wav", **kwargs)
    else:
        torchaudio.save(save_path, waveform, sr, precision=16, channels_first=True, format="wav", **kwargs)


def torchaudio_info(wav_path: str, **kwargs):
    return torchaudio.info(wav_path, **kwargs)


def torchaudio_resample(orig_freq, new_freq, waveform: torch.Tensor) -> torch.Tensor:
    """
    重采样,
    torchaudio.transforms.Resample是一个用于对音频信号进行重采样的类，它可以将一个频率的信号转换为另一个频率的信号12。
    重采样的原理是使用带限插值法，即在原始信号的采样点之间插入新的采样点，以达到目标频率
    重采样的效果是改变了信号的采样率，但不会改变信号的持续时间。例如，如果一个信号的原始频率是16000 Hz，持续时间是10 秒，那么重采样到8000 Hz后，信号的持续时间仍然是10 秒，但采样点的数量减少了一半。
    重采样的目的是为了适应不同的应用场景，例如降低信号的存储空间，提高信号的处理效率，或者匹配不同的设备要求14。
    重采样的缺点是可能会损失信号的一些信息，导致信号的质量下降。为了减少这种损失，重采样的过程中会使用低通滤波器来去除高于目标频率一半的信号分量，以避免混叠现象13。
    :param orig_freq:
    :param new_freq:
    :param waveform:(1, sample_num)
    :return: waveform:(1, sample_num)
    """
    # wav: shape(1,len)
    wav = torchaudio.transforms.Resample(orig_freq, new_freq)(waveform)
    return wav


def torchaudio_respeed(orig_freq: int, factor: float, waveform: torch.Tensor):
    """
    改变音频速度 , 采样率不变, 波形长度变为1/factor
    :param factor:  变更后的速度
    :param orig_freq: 原始采样率
    :param waveform: (1, num_sample)
    :return: waveform: (1, num_sample)
    """
    # wav: shape(1,len), _: None, 传入waveform需要是float
    wav, _ = torchaudio.transforms.Speed(orig_freq=orig_freq, factor=factor)(waveform)
    return wav


def torchaudio_fbank(waveform: torch.Tensor, num_mel_bins=80, frame_length=25, frame_shift=10, dither=0.0,
                     sample_rate=16000):
    """
    计算fbank特征
    :param waveform:  要求使用torchaudio.load(wav_path)得到, 其采用了默认的normal参数, 数据压缩到了float32:[-1.0,1.0], 原始音频多是int16
    :param num_mel_bins:
    :param frame_length:
    :param frame_shift:
    :param dither:
    :param sample_rate:
    :return:
    """
    waveform = waveform * (1 << 15)
    # waveform = waveform.float()
    return torchaudio.compliance.kaldi.fbank(waveform,
                                             num_mel_bins=num_mel_bins,
                                             frame_length=frame_length,
                                             frame_shift=frame_shift,
                                             dither=dither,
                                             sample_frequency=sample_rate,
                                             energy_floor=0.0, )


def torchaudio_mfcc(waveform,
                    num_mel_bins=23, frame_length=25, frame_shift=10, dither=0.0,
                    num_ceps=40, high_freq=0.0, low_freq=20.0, sample_rate=16000, ):
    """
    计算mfcc特征
    :param waveform:
    :param num_mel_bins:
    :param frame_length:
    :param frame_shift:
    :param dither:
    :param sample_rate:
    :param num_ceps:
    :param high_freq:
    :param low_freq:
    :return:
    """
    waveform = waveform * (1 << 15)
    return kaldi.mfcc(waveform,
                      num_mel_bins=num_mel_bins,
                      frame_length=frame_length,
                      frame_shift=frame_shift,
                      dither=dither,
                      num_ceps=num_ceps,
                      high_freq=high_freq,
                      low_freq=low_freq,
                      sample_frequency=sample_rate)


def get_sample_count(audio_file_path: str):
    """
    得到路径所指音频的采样点数
    output->
    sample_count: 采样点数
    sample_rate: 采样率
    """
    return _get_sample_count_wave(audio_file_path)


def _get_sample_count_wave(file_path):
    """比较快"""
    with wave.open(file_path, 'rb') as audio_file:
        sample_count = audio_file.getnframes()
        sample_rate = audio_file.getframerate()
    return sample_count, sample_rate


def _get_sample_count_torchaudio(file_path):
    """比较慢"""
    waveform, sr = torchaudio.load(file_path)
    return len(waveform[0]), sr


def get_padding_mask(seq_q: torch.Tensor, seq_k: torch.Tensor, pad_id: int = 0):
    """
    根据seq_k 中pad_id的位置， 得到seq_q对于seq_k的padding_attn_mask,
    seq_q 只是提供q_len,并不影响mask的分布,
    input->
    seq_q: (batch_size, len_q)
    seq_k: (batch_size, len_k)
    output->
    mask: (batch_size, len_q, len_k)
    """
    batch_size, len_q = seq_q.size()
    batch_size, len_k = seq_k.size()
    mask = seq_k.eq(pad_id)
    mask = mask.unsqueeze(1).expand(batch_size, len_q, len_k)
    return mask


def get_sequence_mask(seq):
    """
    得到一个序列的self-attention的mask
    input->
    seq: [batch_size, seq_len]
    output->
    mask: [batch_size, seq_len, seq_len]
    """
    batch_size, seq_len = seq.size()
    mask = torch.triu(torch.ones((seq_len, seq_len), dtype=torch.uint8, device=seq.device), diagonal=1)
    mask = mask.unsqueeze(0).expand(batch_size, -1, -1)
    return mask


def get_scp_for_wav_dir(wav_dir: str, wav_scp_file_path: str):
    from .utils_file import makedir_for_file, join_path, get_file_pure_name_from_path, get_other_file_in_same_dir
    makedir_for_file(wav_scp_file_path)
    # text_scp = get_other_file_in_same_dir(wav_scp_file_path, 'text.txt.scp')
    wav_path_list = glob.glob(join_path(wav_dir, "*.wav"))
    # with codecs.open(wav_scp_file_path, 'w', encoding='utf-8') as f, open(text_scp, 'w', encoding='utf-8') as f2:
    with codecs.open(wav_scp_file_path, 'w', encoding='utf-8') as f:
        for wav_path in wav_path_list:
            f.write(f"{get_file_pure_name_from_path(wav_path)} {wav_path}\n")
            # f2.write(f"{get_file_pure_name_from_path(wav_path)} 我是耿雪龙焚烧发电萨芬\n")


class WavScpAudioDataset(Dataset):
    def __init__(self, data_file):
        from .utils_file import load_tuple_list_from_scp
        self.items = load_tuple_list_from_scp(data_file)

    def __len__(self):
        return len(self.items)

    def __getitem__(self, idx):
        return self.items[idx]


class CollateFuncForCMVN(object):
    """
    Collate function for AudioDataset
    """

    def __init__(self, feat_dim, resample_rate):
        self.feat_dim = feat_dim
        self.resample_rate = resample_rate
        pass

    def __call__(self, batch):
        mean_stat = torch.zeros(self.feat_dim)
        var_stat = torch.zeros(self.feat_dim)
        number = 0
        for item in batch:
            value = item[1].strip().split(",")
            # len(value) == 3 means segmented wav.scp,->path, start_frame, end_frame
            # len(value) == 1 means original wav.scp
            assert len(value) == 3 or len(value) == 1
            wav_path = value[0]
            sample_rate = torchaudio.info(wav_path).sample_rate
            resample_rate = sample_rate
            if len(value) == 3:
                start_frame = int(float(value[1]) * sample_rate)
                end_frame = int(float(value[2]) * sample_rate)
                waveform, sample_rate = torchaudio.load(
                    filepath=wav_path,
                    num_frames=end_frame - start_frame,
                    frame_offset=start_frame)
            else:
                waveform, sample_rate = torchaudio.load(item[1])

            waveform = waveform * (1 << 15)
            if self.resample_rate != 0 and self.resample_rate != sample_rate:
                resample_rate = self.resample_rate
                waveform = torchaudio.transforms.Resample(
                    orig_freq=sample_rate, new_freq=resample_rate)(waveform)

            mat = kaldi.fbank(waveform,
                              num_mel_bins=self.feat_dim,
                              dither=0.0,
                              energy_floor=0.0,
                              sample_frequency=resample_rate)
            mean_stat += torch.sum(mat, dim=0)
            var_stat += torch.sum(torch.square(mat), dim=0)
            number += mat.shape[0]
        return number, mean_stat, var_stat


class GxlStringBuffer:
    def __init__(self):
        self.strings = ""

    @property
    def linefeed(self):
        if platform.system().lower() == 'windows':
            return '\r\n'
        else:
            return '\n'

    def print(self, string_data: str, end=None):
        if end is None:
            end = self.linefeed
        self.strings += (string_data + end)

    def to_string(self):
        """
        返回缓冲区的字符串
        :return:
        """
        return self.strings

    def to_list(self):
        """
        末尾带有换行符
        :return:
        """
        return [str_i + self.linefeed for str_i in self.to_list_clean()]

    def to_list_clean(self):
        """
        末尾不带换行符
        :return:
        """
        return self.strings.split(self.linefeed)

    def __str__(self):
        return self.to_string()


def do_compute_cmvn_stats(num_workers: int, config_yaml_path: str,
                          in_wav_scp_path: str, out_cmvn_json_path: str,
                          log_interval: int = 1000):
    """
    compute cmvn stats
    config_yaml内容要求:
    dataset_conf:
        resample_conf:
            resample_rate: 16000 -------->采样率
        fbank_conf:
            num_mel_bins: 80------>特征维度

    :param num_workers:
    :param config_yaml_path:
    :param in_wav_scp_path:
    :param out_cmvn_json_path:
    :param log_interval:
    :return:
    """
    from .utils_file import makedir_for_file
    configs = GxlNode.get_config_from_yaml(config_yaml_path)
    feat_dim = configs.dataset_conf.fbank_conf.num_mel_bins
    makedir_for_file(out_cmvn_json_path)
    resample_rate = 0
    if 'resample_conf' in configs['dataset_conf']:
        resample_rate = configs['dataset_conf']['resample_conf']['resample_rate']
        print('using resample and new sample rate is {}'.format(resample_rate))
    collate_func = CollateFuncForCMVN(feat_dim, resample_rate)
    dataset = WavScpAudioDataset(in_wav_scp_path)
    batch_size = 20
    data_loader = DataLoader(dataset,
                             batch_size=batch_size,
                             shuffle=True,
                             sampler=None,
                             num_workers=num_workers,
                             collate_fn=collate_func)
    with torch.no_grad():
        all_number = 0
        all_mean_stat = torch.zeros(feat_dim)
        all_var_stat = torch.zeros(feat_dim)
        wav_number = 0
        for i, batch in enumerate(data_loader):
            number, mean_stat, var_stat = batch
            all_mean_stat += mean_stat
            all_var_stat += var_stat
            all_number += number
            wav_number += batch_size

            if wav_number % log_interval == 0:
                print(f'processed {wav_number} wavs, {all_number} frames',
                      file=sys.stderr,
                      flush=True)
    cmvn_info = {
        'mean_stat': list(all_mean_stat.tolist()),
        'var_stat': list(all_var_stat.tolist()),
        'frame_num': all_number
    }
    from .utils_file import write_dict_to_json
    write_dict_to_json(cmvn_info, out_cmvn_json_path)


def do_text2token(text_scp_path: str, output_path: str,
                  nchar: int = 1, skip_ncols: int = 1,
                  space: str = '<space>', bpe_model_path: str = None,
                  non_lang_syms_path: str = None,
                  trans_type: str = 'char',
                  ):
    """
    do text.txt to token
    :param output_path:
    :param nchar:number of characters to split, i.e., \
                        aabb -> a a b b with -n 1 and aa bb with -n 2
    :param skip_ncols:skip first n columns, 是跳过处理, 直接保留,而不是放弃
    :param space:space symbol
    :param bpe_model_path:bpe model for english part
    :param non_lang_syms_path:list of non-linguistic symobles,
                         e.g., <NOISE> etc.
    :param text_scp_path:input text.txt
    :param trans_type:  choices=["char", "phn", "cn_char_en_bpe"],
                        help=Transcript type. char/phn. e.g., for TIMIT
                             FADG0_SI1279 -
                             If trans_type is char, read from
                             SI1279.WRD file -> "bricks are an alternative"
                             Else if trans_type is phn,
                             read from SI1279.PHN file ->
                             "sil b r ih sil k s aa r er n aa l
                             sil t er n ih sil t ih v sil)
    :return:
    """

    def exist_or_not(i, match_pos):
        start_pos = None
        end_pos = None
        for pos in match_pos:
            if pos[0] <= i < pos[1]:
                start_pos = pos[0]
                end_pos = pos[1]
                break
        return start_pos, end_pos

    def seg_char(sent):
        """
        返回汉字和分割的字符
        :param sent:
        :return:
        """
        pattern = re.compile(r'([\u4e00-\u9fa5])')
        # 返回汉字和分割是字符
        chars = pattern.split(sent)
        chars = [w for w in chars if len(w.strip()) > 0]
        return chars

    rs = []
    if non_lang_syms_path is not None:
        from .utils_file import load_list_file
        nls = load_list_file(non_lang_syms_path)
        rs = [re.compile(re.escape(x)) for x in nls]
    sp = None
    if bpe_model_path is not None:
        import sentencepiece as spm
        sp = spm.SentencePieceProcessor()
        sp.load(bpe_model_path)

    if text_scp_path is not None:
        f = codecs.open(text_scp_path, encoding="utf-8")
    else:
        f = codecs.getreader("utf-8")(
            sys.stdin if is_python2 else sys.stdin.buffer)
    buffer = GxlStringBuffer()
    line = f.readline()
    n = nchar
    while line:
        x = line.split()
        buffer.print(' '.join(x[:skip_ncols]), end=" ")
        a = ' '.join(x[skip_ncols:])

        # get all matched positions
        match_pos = []
        for r in rs:
            i = 0
            while i >= 0:
                m = r.search(a, i)
                if m:
                    match_pos.append([m.start(), m.end()])
                    i = m.end()
                else:
                    break

        if len(match_pos) > 0:
            #  如果有特殊符号, 那么只能支持char分类
            chars = []
            i = 0
            while i < len(a):
                start_pos, end_pos = exist_or_not(i, match_pos)
                if start_pos is not None:
                    chars.append(a[start_pos:end_pos])
                    i = end_pos
                else:
                    chars.append(a[i])
                    i += 1
            a = chars

        if trans_type == "phn":
            a = a.split(" ")
        elif trans_type == "cn_char_en_bpe":
            b = seg_char(a)
            a = []
            for j in b:
                # we use "▁" to instead of blanks among english words
                # warning: here is "▁", not "_"
                for l in j.strip().split("▁"):
                    if not l.encode('UTF-8').isalpha():
                        a.append(l)
                    else:
                        if sp is None:
                            raise Exception("please set bpe_model_path when trans_type is cn_char_en_bpe")
                        for k in sp.encode_as_pieces(l):
                            a.append(k)
        else:
            a = [a[j:j + n] for j in range(0, len(a), n)]

        a_flat = []
        # 确保都是字符串
        for z in a:
            a_flat.append("".join(z))

        a_chars = [z.replace(' ', space) for z in a_flat]
        if trans_type == "phn":
            a_chars = [z.replace("sil", space) for z in a_chars]
        buffer.print(' '.join(a_chars))
        line = f.readline()
    lines = buffer.to_list_clean()
    cut_result = [line.split()[1:] for line in lines]
    # 将列表展平，并用换行符连接
    tr_result = "\n".join("\n".join(row) for row in cut_result)
    # 对行进行排序
    sort_result = "\n".join(sorted(tr_result.split("\n")))
    # 去除重复行
    uniq_result = "\n".join(sorted(set(sort_result.split("\n"))))
    # 过滤掉只包含空白字符的行
    grep_result = "\n".join(line for line in uniq_result.split("\n") if line.strip())
    # 在每一行末尾添加数字
    items = grep_result.split("\n")
    awk_result = "\n".join(f"{line} {i + 1}" for i, line in enumerate(items))
    awk_result += f"\n<sos/eos> {len(items) + 1}\n"
    from .utils_file import makedir_for_file
    makedir_for_file(output_path)
    # 将结果追加到文件
    with open(output_path, "w", encoding='utf-8') as f:
        f.write(awk_result)


def do_make_raw_list(wav_scp_path: str, text_scp_path: str,
                     output_jsonl_path, segments_path: str = None):
    """
    do make raw list
    将wav_scp text_scp合并成一个jsonl:data.list
    :param wav_scp_path:
    :param text_scp_path:
    :param output_jsonl_path:
    :param segments_path:
    :return:
    """
    from .utils_file import load_dict_from_scp, write_dict_list_to_jsonl, makedir_for_file
    makedir_for_file(output_jsonl_path)
    wav_dict = load_dict_from_scp(wav_scp_path)
    text_dict = load_dict_from_scp(text_scp_path)
    segments_dict = {}
    if segments_path is not None:
        with open(segments_path, 'r', encoding='utf8') as fin:
            for line in fin:
                arr = line.strip().split()
                assert len(arr) == 4
                segments_dict[arr[0]] = (arr[1], float(arr[2]), float(arr[3]))
    json_list = []
    for key, txt in text_dict.items():
        if segments_path is None:
            if key not in wav_dict:
                print('warning from gxl: {} in text_scp is not in wav_dict'.format(key))
                continue
            wav_path = wav_dict[key]
            json_line = dict(key=key, wav=wav_path, txt=txt)
        else:
            if key not in segments_dict:
                print('warning from gxl: {} in text_scp is not in segments_dict'.format(key))
                continue
            wav_key = segments_dict[key][0]
            start_time = segments_dict[key][1]
            end_time = segments_dict[key][2]
            if wav_key not in wav_dict:
                print(f'warning from gxl: {wav_key} in segments_dict[{key}][0] is not in wav_dict')
                continue
            wav_path = wav_dict[wav_key]
            json_line = dict(key=key, wav=wav_path, txt=txt, start_time=start_time, end_time=end_time)
        json_list.append(json_line)
    write_dict_list_to_jsonl(json_list, output_jsonl_path)
def do_change_file_suffix(tar_file_path, param):
    str_1 = tar_file_path.split('.')[:-1]
    return '.'.join(str_1) + '.' + param

def write_to_tar_file(data_list: list[tuple], tar_file_path: str, resample=16000, i=-1):
    """
    将数据写入tar文件，
    data_list: item: (key, text.txt, wav_path)
    """
    print(f'开始处理第{i}个shard')
    AUDIO_FORMAT_SETS = {'flac', 'mp3', 'm4a', 'ogg', 'opus', 'wav', 'wma'}
    from .utils_file import makedir_for_file
    makedir_for_file(tar_file_path)
    finished_path = do_change_file_suffix(tar_file_path, 'finished')
    with tarfile.open(tar_file_path, "w") as tar:
        for item in tqdm(data_list, total=len(data_list), desc=f"shard_{i}"):
            key, txt, wav = item
            suffix = wav.split('.')[-1]
            assert suffix in AUDIO_FORMAT_SETS, f"不支持的音频格式{suffix},仅支持{AUDIO_FORMAT_SETS}"
            # read & resample
            audio, sample_rate = torchaudio.load(wav, normalize=False)
            if sample_rate != resample:
                audio = torchaudio.transforms.Resample(
                    sample_rate, resample)(audio.float())
                audio = audio.to(torch.int16)
            # change format to wav
            f = io.BytesIO()
            torchaudio.save(f, audio, resample, format="wav", bits_per_sample=16)
            suffix = "wav"
            f.seek(0)
            data = f.read()
            assert isinstance(txt, str), f"txt必须是str类型"
            txt_file_name = key + '.txt'
            txt = txt.encode('utf8')
            txt_data = io.BytesIO(txt)
            txt_info = tarfile.TarInfo(txt_file_name)
            txt_info.size = len(txt)
            tar.addfile(txt_info, txt_data)

            wav_file = key + '.' + suffix
            wav_data = io.BytesIO(data)
            wav_info = tarfile.TarInfo(wav_file)
            wav_info.size = len(data)
            tar.addfile(wav_info, wav_data)
    print(f'第{i}个shard处理完成')
    with open(finished_path, 'w') as f:
        pass


def do_make_shard_file(wav_scp_file_path: str, text_scp_file_path: str, output_dir: str, num_utt_per_shard: int = 1000,
                       num_threads=32, prefix_for_tar_file: str = "shard", resample: int = 16000,
                       ):
    """
    得到一个shard文件组成的目录, logger must is not None
    """
    logging_print('开始打shard for ' + prefix_for_tar_file)
    logging_print('wav_scp: ' + wav_scp_file_path)
    logging_print('text_scp: ' + text_scp_file_path)
    from .utils_file import load_dict_from_scp
    wav_dic = load_dict_from_scp(wav_scp_file_path)
    data = []
    text_dic = load_dict_from_scp(text_scp_file_path)
    for k, text in text_dic.items():
        if k not in wav_dic:
            logging_print(f"warning: {k}不在wav_scp文件中")
            continue
        data.append((k, text, wav_dic[k]))
    logging_print(f"共有{len(data)}个utt")
    chunks = [data[i:i + num_utt_per_shard] for i in range(0, len(data), num_utt_per_shard)]
    os.makedirs(output_dir, exist_ok=True)
    logging_print(f"共有{len(chunks)}个shard")
    # Using thread pool to speedup
    pool = multiprocessing.Pool(processes=num_threads)
    shards_list = []
    for i, chunk in enumerate(chunks):
        tar_file_path = os.path.join(output_dir,
                                     '{}_{:09d}.tar'.format(prefix_for_tar_file, i))
        shards_list.append(tar_file_path)
        finished_file_path = do_change_file_suffix(tar_file_path, 'finished')
        if os.path.exists(finished_file_path):
            continue
        pool.apply_async(
            write_to_tar_file,
            (chunk, tar_file_path, resample, i))

    pool.close()
    pool.join()
    logging_print('打shard结束, 保存shard列表')
    with open(os.path.join(output_dir, 'shards_list.txt'), 'w', encoding='utf8') as fout:
        for name in shards_list:
            fout.write(name + '\n')
    logging_print('打shard完全结束')


def do_remove_blank_for_text_scp(text_scp_path: str, ):
    """
    删除text_scp中的空白行
    """
    dir_name = os.path.dirname(text_scp_path)
    temp_file = os.path.join(dir_name, 'temp.txt')
    with open(text_scp_path, 'r', encoding='utf8') as fin, open(temp_file, 'w', encoding='utf8') as fout:
        for line in fin:
            items = line.strip().split()
            key = items[0]
            txt = "".join(items[1:])
            fout.write(f"{key} {txt}\n")
    from .utils_file import load_dict_from_scp, write_dict_to_scp
    dic = load_dict_from_scp(temp_file)
    os.remove(temp_file)
    write_dict_to_scp(dic, text_scp_path)


def do_dict2simpleNamespaceObj(dict_obj: dict):
    """
    将一个字典转换为命名空间对象,
    命名空间对象可以修改key对应的value值
    可以通过.的方式调用键值对用的value值,如果调用没设置的键值,则直接报错,
    :param dict_obj:
    :return:
    """
    return types.SimpleNamespace(**dict_obj)



import os.path
import re

from torch.utils.tensorboard import SummaryWriter

from ..config.gxl_config import GxlNode
from ..utils import utils_file
from ..gxl_model_wenet.utils.train_utils import *
from ..gxl_dataset_wenet.gxl_dataset import GxlAsrDataset
from ..gxl_model_wenet.init_model import init_model, ASRModel
from ..gxl_model_wenet.utils.executor import Executor
from ..gxl_model_wenet.utils.checkpoint import load_checkpoint, save_checkpoint, load_trained_modules
from ..gxl_lr_scheduler_wenet.scheduler import WarmupLR, NoamHoldAnnealing
from ..utils import utils_data

import torch
import torch.nn as nn
import torch.optim as optim
import torch.distributed as dist
import torch.multiprocessing as mp
from torch.nn.parallel import DistributedDataParallel as DDP


def read_non_lang_symbols(non_lang_sym_path):
    """read non-linguistic symbol from file.
    The file format is like below:
    {NOISE}\n
    {BRK}\n
    ...
    Args:
        non_lang_sym_path: non-linguistic symbol file path, None means no any
        syms.
    """
    if non_lang_sym_path is None:
        return None
    else:
        syms = utils_file.load_list_file_clean(non_lang_sym_path)
        non_lang_syms_pattern = re.compile(r"(\[[^\[\]]+\]|<[^<>]+>|{[^{}]+})")
        for sym in syms:
            if non_lang_syms_pattern.fullmatch(sym) is None:
                class BadSymbolFormat(Exception):
                    pass

                raise BadSymbolFormat(
                    "Non-linguistic symbols should be "
                    "formatted in {xxx}/<xxx>/[xxx], consider"
                    " modify '%s' to meet the requirment. "
                    "More details can be found in discussions here : "
                    "https://github.com/wenet-e2e/wenet/pull/819" % (sym))
        return syms


def read_symbol_table(symbol_table_file):
    symbol_table = {}
    with open(symbol_table_file, 'r', encoding='utf8') as fin:
        for line in fin:
            arr = line.strip().split()
            assert len(arr) == 2
            symbol_table[arr[0]] = int(arr[1])
    return symbol_table


class ToyModel(nn.Module):
    def __init__(self):
        super(ToyModel, self).__init__()
        self.conv = nn.Conv2d(3, 16, 3)
        self.flatten = nn.Flatten()
        self.linear = nn.Linear(30 * 30 * 16, 10)
        self.softmax = nn.Softmax(dim=1)

    def forward(self, x):
        x = self.conv(x)
        x = self.flatten(x)
        x = self.linear(x)
        x = self.softmax(x)
        return x


# 定义一个函数，用于在每个进程中进行DDP的初始化、模型的构造、训练的循环等操作
def run(rank, world_size):
    # DDP backend初始化
    dist.init_process_group(backend='nccl')
    # 设置当前进程使用的GPU设备
    torch.cuda.set_device(rank)
    # 构造模型，并将其封装为DDP模型
    model = ToyModel().to(rank)
    ddp_model = DDP(model, device_ids=[rank])
    # 定义损失函数和优化器
    loss_fn = nn.CrossEntropyLoss()
    optimizer = optim.SGD(ddp_model.parameters(), lr=0.001)
    # 模拟一个batch的数据和标签
    inputs = torch.randn(20, 3, 32, 32).to(rank)
    labels = torch.randint(0, 10, (20,)).to(rank)
    # 前向传播
    outputs = ddp_model(inputs)
    # 计算损失
    loss = loss_fn(outputs, labels)
    # 后向传播
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()
    # 打印损失值
    print(f'Rank {rank}, loss: {loss.item()}')


class GxlTrainer(object):
    def __init__(self, train_config_yaml_path: str = os.path.dirname(os.path.abspath(__file__)) + '/train_config.yaml'):
        self.args = GxlNode.get_config_from_yaml(train_config_yaml_path)
        torch.manual_seed(777)
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s %(levelname)s %(message)s')
        self.configs = GxlNode.get_dict_from_yaml(self.args.content_config)

    def prepare_data(self, data_dir: str):
        """
        数据处理：
        1. 移除text_scp中的空白间隔，如果没有空白间隔，则不作为
        2. 根据配置文件中的设置选择是否计算cmvn
        3. 提取训练集和验证集文本中的token并得到token_symbol_table文件
        4. 最后得到dataset需要的data.list数据。
        :param train_wav_scp_path:
        :param train_text_scp_path:
        :param cv_wav_scp_path:
        :param cv_text_scp_path:
        :return:
        """
        train_text_scp_path = os.path.join(data_dir, 'train', 'text.txt')
        train_wav_scp_path = os.path.join(data_dir, 'train', 'wav.scp')
        cv_text_scp_path = os.path.join(data_dir, 'dev', 'text.txt')
        cv_wav_scp_path = os.path.join(data_dir, 'dev', 'wav.scp')
        utils_data.do_remove_blank_for_text_scp(train_text_scp_path)
        utils_data.do_remove_blank_for_text_scp(cv_text_scp_path)
        # if self.args.cmvn is not None:
        #     utils_data.do_compute_cmvn_stats(self.args.num_workers, self.args.content_config,
        #                                      train_wav_scp_path, self.args.cmvn)
        utils_data.do_text2token(train_text_scp_path, self.args.symbol_table)

        utils_data.do_make_raw_list(train_wav_scp_path, train_text_scp_path, self.args.train_data)
        utils_data.do_make_raw_list(cv_wav_scp_path, cv_text_scp_path, self.args.cv_data)

    def train_run(self):
        """"""
        args = self.args
        configs = self.configs
        is_distributed = torch.cuda.is_available() and torch.cuda.device_count() >= 1
        # Init env for ddp OR deepspeed
        if is_distributed:
            world_size, local_rank, rank = init_distributed(args)
        else:
            world_size, local_rank, rank = 1, 0, 0

        # Get dataset & dataloader
        train_dataset, cv_dataset, train_data_loader, cv_data_loader = \
            init_dataset_and_dataloader(args, configs)

        # Do some sanity checks and save config to arsg.model_dir
        configs = check_modify_and_save_config(args, configs)

        # Init asr model from configs
        model, configs = init_model(args, configs)

        # Check model is jitable & print model archtectures
        trace_and_print_model(args, model)

        # Tensorboard summary
        writer = init_summarywriter(args)

        # Dispatch model from cpu to gpu
        if is_distributed:
            model, device = wrap_cuda_model(args, model)

        # Get optimizer & scheduler
        model, optimizer, scheduler = init_optimizer_and_scheduler(
            args, configs, model)

        # Save checkpoints
        save_model(model, info_dict={
            "save_time": datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S'),
            "tag": "init", **configs
        })

        # Get executor
        executor = Executor()
        executor.step = configs["init_infos"].get('step', 0)

        # Init scaler, used for pytorch amp mixed precision training
        scaler = None
        if args.use_amp:
            scaler = torch.cuda.amp.GradScaler()

        # Start training loop
        start_epoch = configs["init_infos"].get('epoch', -1) + 1
        final_epoch = None
        for epoch in range(start_epoch, configs.get('max_epoch', 100)):
            train_dataset.set_epoch(epoch)
            configs['epoch'] = epoch

            lr = optimizer.param_groups[0]['lr']
            logging.info('Epoch {} TRAIN info lr {} rank {}'.format(epoch, lr, rank))

            # NOTE(xcsong): Why we need a new group? see `train_utils.py::wenet_join`
            if is_distributed:
                group_join = dist.new_group(backend="gloo",
                                        timeout=datetime.timedelta(seconds=args.timeout))
            else:
                group_join = None
            if is_distributed:
                dist.barrier()  # NOTE(xcsong): Ensure all ranks start Train at the same time.
            executor.train(model, optimizer, scheduler, train_data_loader,
                           writer, configs, scaler, group_join)
            if is_distributed:
                dist.barrier()  # NOTE(xcsong): Ensure all ranks start CV at the same time.
            total_loss, num_seen_utts = executor.cv(model, cv_data_loader, configs)
            cv_loss = total_loss / num_seen_utts

            logging.info('Epoch {} CV info cv_loss {} rank {}'.format(epoch, cv_loss, rank))
            info_dict = {
                'epoch': epoch, 'lr': lr, 'cv_loss': cv_loss, 'step': executor.step,
                'save_time': datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S'),
                'tag': str(epoch), **configs
            }
            log_per_epoch(writer, info_dict=info_dict)
            save_model(model, info_dict=info_dict)

            final_epoch = epoch

        if final_epoch is not None and rank == 0:
            final_model_path = os.path.join(args.model_dir, 'final.pt')
            os.remove(final_model_path) if os.path.exists(final_model_path) else None
            os.symlink('{}.pt'.format(final_epoch), final_model_path)
            writer.close()

    def run_multi_thread(self):
        pass

    @staticmethod
    def run_multi_thread_test():
        # 获取GPU数量
        os.environ['CUDA_VISIBLE_DEVICES'] = '1,2,3'
        os.environ['TRITON_LOG_LEVEL'] = "0"
        os.environ['RANK'] = "0"
        os.environ['WORLD_SIZE'] = "3"
        os.environ['MASTER_ADDR'] = "127.0.0.1"
        os.environ['MASTER_PORT'] = "100862"
        n_gpus = torch.cuda.device_count()
        # 启动多个进程，每个进程调用run函数
        mp.spawn(run, args=(n_gpus,), nprocs=n_gpus, join=True)



import torch
from torch import nn

from .competition.multi_head_attention import MultiHeadAttention
from .competition.positional_encoding import PositionalEncoding
from .competition.positional_wise_feed_forward import PositionWiseFeedForward

from .utils.padding_mask import padding_mask
from .utils.sequence_mask import sequence_mask


class DecoderLayer(nn.Module):
    def __init__(self, model_dim: int, num_heads: int = 8, ffn_dim: int = 2048, dropout: float = 0.0):
        super(DecoderLayer, self).__init__()
        self.self_attention = MultiHeadAttention(model_dim, num_heads, dropout)
        self.context_attention = MultiHeadAttention(model_dim, num_heads, dropout)
        self.feed_forward = PositionWiseFeedForward(model_dim, ffn_dim, dropout)

    def forward(self, decoder_input, encoder_output,
                self_attn_mask=None, context_attn_mask=None):
        decoder_output, self_attention = self.self_attention(decoder_input, decoder_input,
                                                             decoder_input, self_attn_mask)
        decoder_output, context_attention = self.context_attention(decoder_output, encoder_output,
                                                                   encoder_output, context_attn_mask)
        decoder_output = self.feed_forward(decoder_output)
        return decoder_output, self_attention, context_attention


class Decoder(nn.Module):
    def __init__(self, vocab_size: int, max_seq_len: int,
                 num_layers: int = 6, model_dim: int = 512, num_heads: int = 8, ffn_dim: int = 2048,
                 dropout: float = 0.0, padding_idx: int = 0):
        super(Decoder, self).__init__()
        self.seq_embedding = nn.Embedding(vocab_size, model_dim, padding_idx=padding_idx)
        self.pos_embedding = PositionalEncoding(model_dim, max_seq_len)

        self.num_layers = num_layers

        self.decoder_layers = nn.ModuleList(
            [DecoderLayer(model_dim, num_heads, ffn_dim, dropout) for _ in range(num_layers)])

    def forward(self, input_data, input_lens, encoder_output, context_attn_mask):
        """
        input->
        input_data: (batch_size, seq_len1)
        input_lens: (batch_size, 1)
        encoder_output: (batch_size, seq_len2, model_dim)
        context_attn_mask: (batch_size, seq_len1, seq_len2)
        """
        self_padding_mask = padding_mask(input_data, input_data)
        self_seq_mask = sequence_mask(input_data)
        self_attn_mask = torch.gt(self_seq_mask + self_padding_mask, 0)
        output = self.seq_embedding(input_data)
        output += self.pos_embedding(input_lens)
        self_attentions = []
        context_attentions = []
        for layer in self.decoder_layers:
            output, self_attn, context_attn = layer(output, encoder_output, self_attn_mask, context_attn_mask)
            self_attentions.append(self_attn)
            context_attentions.append(context_attn)
        return output, self_attentions, context_attentions



from torch import nn
from torchvision import models
from torchvision.models import ResNet18_Weights

from gxl_ai_utils import utils
from gxl_ai_utils import AiConstant

class RNNGxlModel(nn.Module):
    def __init__(self, output_num=102, feature_extract=True):
        super(RNNGxlModel, self).__init__()
        self.resnet_block = models.resnet18(weights=ResNet18_Weights.DEFAULT)
        utils.utils_model.set_parameter_requires_grad(self.resnet_block, feature_extract)
        fc_in = self.resnet_block.fc.in_features
        self.resnet_block.fc = nn.Linear(fc_in, output_num)

    def forward(self, X):
        return self.resnet_block(X)


# class RnnModelConfig:
#     def __init__(self):
#         self.data_path = AiConstant.DATA_PATH +'text_data'
#         self.emb_n = 300
#         self.vocab_size = 30000
#         self.dropout = 0.001
#         self.class_list = [x.strip() for x in open(
#             self.data_path + '/class.txt').readlines()]
#         self.num_classes = len(self.class_list)
#         self.hidden_n = 333
#         self.layers = 3


# class RNNTextClassifyModel(nn.Module):
#     def __init__(self, config=RnnModelConfig()):
#         super(RNNTextClassifyModel, self).__init__()
#         self.config = config
#         self.embedding = nn.Embedding(self.config.vocab_size, self.config.emb_n, padding_idx=self.config.vocab_size - 1)
#         self.lstm = nn.LSTM(self.config.emb_n, self.config.hidden_n, self.config.layers, bidirectional=True,
#                             batch_first=True, dropout=self.config.dropout)
#         self.fc = nn.Linear(self.config.hidden_n * 2, self.config.num_classes)

#     def forward(self, X):
#         """"""
#         out = self.embedding(X)
#         out = self.lstm(out)[0]
#         out = self.fc(out[:, -1, :])
#         return out


def test_rnn_model():
    train_loader, valid_loader, test_loader = utils_text.provide_loader()
    model = RNNTextClassifyModel()
    one_batch_X, one_batch_lens, one_batch_y = next(iter(train_loader))
    model.eval()
    model((one_batch_X, one_batch_lens))


from typing import List, Tuple

import torch
from ...utils.common import log_add


class Sequence():

    __slots__ = {'hyp', 'score', 'cache'}

    def __init__(
        self,
        hyp: List[torch.Tensor],
        score,
        cache: List[torch.Tensor],
    ):
        self.hyp = hyp
        self.score = score
        self.cache = cache


class PrefixBeamSearch():

    def __init__(self, encoder, predictor, joint, ctc, blank):
        self.encoder = encoder
        self.predictor = predictor
        self.joint = joint
        self.ctc = ctc
        self.blank = blank

    def forward_decoder_one_step(
            self, encoder_x: torch.Tensor, pre_t: torch.Tensor,
            cache: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        padding = torch.zeros(pre_t.size(0), 1, device=encoder_x.device)
        pre_t, new_cache = self.predictor.forward_step(pre_t.unsqueeze(-1),
                                                       padding, cache)
        x = self.joint(encoder_x, pre_t)  # [beam, 1, 1, vocab]
        x = x.log_softmax(dim=-1)
        return x, new_cache

    def prefix_beam_search(self,
                           speech: torch.Tensor,
                           speech_lengths: torch.Tensor,
                           decoding_chunk_size: int = -1,
                           beam_size: int = 5,
                           num_decoding_left_chunks: int = -1,
                           simulate_streaming: bool = False,
                           ctc_weight: float = 0.3,
                           transducer_weight: float = 0.7):
        """prefix beam search
           also see wenet.transducer.transducer.beam_search
        """
        assert speech.shape[0] == speech_lengths.shape[0]
        assert decoding_chunk_size != 0
        device = speech.device
        batch_size = speech.shape[0]
        assert batch_size == 1

        # 1. Encoder
        encoder_out, _ = self.encoder(
            speech, speech_lengths, decoding_chunk_size,
            num_decoding_left_chunks)  # (B, maxlen, encoder_dim)
        maxlen = encoder_out.size(1)

        ctc_probs = self.ctc.log_softmax(encoder_out).squeeze(0)
        beam_init: List[Sequence] = []

        # 2. init beam using Sequence to save beam unit
        cache = self.predictor.init_state(1, method="zero", device=device)
        beam_init.append(Sequence(hyp=[self.blank], score=0.0, cache=cache))
        # 3. start decoding (notice: we use breathwise first searching)
        # !!!! In this decoding method: one frame do not output multi units. !!!!
        # !!!!    Experiments show that this strategy has little impact      !!!!
        for i in range(maxlen):
            # 3.1 building input
            # decoder taking the last token to predict the next token
            input_hyp = [s.hyp[-1] for s in beam_init]
            input_hyp_tensor = torch.tensor(input_hyp,
                                            dtype=torch.int,
                                            device=device)
            # building statement from beam
            cache_batch = self.predictor.cache_to_batch(
                [s.cache for s in beam_init])
            # build score tensor to do torch.add() function
            scores = torch.tensor([s.score for s in beam_init]).to(device)

            # 3.2 forward decoder
            logp, new_cache = self.forward_decoder_one_step(
                encoder_out[:, i, :].unsqueeze(1),
                input_hyp_tensor,
                cache_batch,
            )  # logp: (N, 1, 1, vocab_size)
            logp = logp.squeeze(1).squeeze(1)  # logp: (N, vocab_size)
            new_cache = self.predictor.batch_to_cache(new_cache)

            # 3.3 shallow fusion for transducer score
            #     and ctc score where we can also add the LM score
            logp = torch.log(
                torch.add(transducer_weight * torch.exp(logp),
                          ctc_weight * torch.exp(ctc_probs[i].unsqueeze(0))))

            # 3.4 first beam prune
            top_k_logp, top_k_index = logp.topk(beam_size)  # (N, N)
            scores = torch.add(scores.unsqueeze(1), top_k_logp)

            # 3.5 generate new beam (N*N)
            beam_A = []
            for j in range(len(beam_init)):
                # update seq
                base_seq = beam_init[j]
                for t in range(beam_size):
                    # blank: only update the score
                    if top_k_index[j, t] == self.blank:
                        new_seq = Sequence(hyp=base_seq.hyp.copy(),
                                           score=scores[j, t].item(),
                                           cache=base_seq.cache)

                        beam_A.append(new_seq)
                    # other unit: update hyp score statement and last
                    else:
                        hyp_new = base_seq.hyp.copy()
                        hyp_new.append(top_k_index[j, t].item())
                        new_seq = Sequence(hyp=hyp_new,
                                           score=scores[j, t].item(),
                                           cache=new_cache[j])
                        beam_A.append(new_seq)

            # 3.6 prefix fusion
            fusion_A = [beam_A[0]]
            for j in range(1, len(beam_A)):
                s1 = beam_A[j]
                if_do_append = True
                for t in range(len(fusion_A)):
                    # notice: A_ can not fusion with A
                    if s1.hyp == fusion_A[t].hyp:
                        fusion_A[t].score = log_add(
                            [fusion_A[t].score, s1.score])
                        if_do_append = False
                        break
                if if_do_append:
                    fusion_A.append(s1)

            # 4. second pruned
            fusion_A.sort(key=lambda x: x.score, reverse=True)
            beam_init = fusion_A[:beam_size]

        return beam_init, encoder_out



import torch
from torch import nn
import numpy as np
import torch.nn.functional as F


def padding_mask(seq_q: torch.Tensor, seq_k: torch.Tensor, pad_id: int = 0):
    """
    根据seq_k 中pad_id的位置， 得到seq_q对于seq_k的padding_attn_mask,
    seq_q 只是提供q_len,并不影响mask的分布,
    input->
    seq_q: (batch_size, len_q)
    seq_k: (batch_size, len_k)
    output->
    mask: (batch_size, len_q, len_k)
    """
    batch_size, len_q = seq_q.size()
    batch_size, len_k = seq_k.size()
    mask = seq_k.eq(pad_id)
    mask = mask.unsqueeze(1).expand(batch_size, len_q, len_k)
    return mask


def sequence_mask(seq):
    """
    得到一个序列的self-attention的mask
    input->
    seq: [batch_size, seq_len]
    output->
    mask: [batch_size, seq_len, seq_len]
    """
    batch_size, seq_len = seq.size()
    mask = torch.triu(torch.ones((seq_len, seq_len), dtype=torch.uint8, device=seq.device), diagonal=1)
    mask = mask.unsqueeze(0).expand(batch_size, -1, -1)
    return mask


class ScaledDotProductAttention(nn.Module):
    """
    Scaled Dot-Product Attention
    缩放 点积 注意力
    """

    def __init__(self, attention_dropout=0.0):
        super().__init__()
        self.softmax = nn.Softmax(dim=2)
        self.dropout = nn.Dropout(attention_dropout)

    def forward(self, q, k, v, scale=None, attn_mask=None):
        """
        q: query , (B,h, T_q, D_q)
        k: key    , (B,h, T_k, D_k)
        v: value  , (B,h, T_v, D_v), 这里在实现上认为D_q=D_k, T_k=T_v
        scale: 缩放因子, int
        attn_mask: (B,h, T_q, T_k), bool, 在需要遮蔽的地方设为true
        return:
            context: (B,h, T_q, D_v),
            attention: (B,h, T_q, T_k)
        """
        attention = torch.matmul(q, k.transpose(-2, -1))
        if scale:
            attention = scale * attention
        if attn_mask is not None:
            attention = attention.masked_fill_(attn_mask, -np.inf)
        attention = self.softmax(attention)
        attention = self.dropout(attention)
        context = torch.matmul(attention, v)
        return context, attention


class MultiHeadAttention(nn.Module):
    def __init__(self, model_dim=512, num_heads=8, dropout=0.0):
        super(MultiHeadAttention, self).__init__()
        self.dim_per_head = model_dim // num_heads
        self.num_heads = num_heads

        self.query = nn.Linear(model_dim, model_dim)
        self.key = nn.Linear(model_dim, model_dim)
        self.value = nn.Linear(model_dim, model_dim)
        self.out = nn.Linear(model_dim, model_dim)

        self.attn = ScaledDotProductAttention(dropout)

        self.dropout = nn.Dropout(dropout)
        self.layer_norm = nn.LayerNorm(model_dim)

    def forward(self, query, key, value, attn_mask=None):
        """
        query: [batch_size, query_len, model_dim]
        key: [batch_size, key_len, model_dim]
        value: [batch_size, value_len, model_dim]
        attn_mask: [batch_size, query_len, key_len]
        """
        residual = query
        batch_size = query.size(0)
        num_heads = self.num_heads
        dim_per_head = self.dim_per_head

        query = self.query(query)
        key = self.key(key)
        value = self.value(value)
        query = query.view(batch_size, -1, num_heads, dim_per_head).transpose(1, 2)
        key = key.view(batch_size, -1, num_heads, dim_per_head).transpose(1, 2)
        value = value.view(batch_size, -1, num_heads, dim_per_head).transpose(1, 2)
        if attn_mask is not None:
            attn_mask = attn_mask.unsqueeze(1)
            attn_mask = attn_mask.repeat(1, num_heads, 1, 1)
        scale = key.size(-1) ** -0.5
        context, attention = self.attn(query, key, value, scale, attn_mask)
        context = context.transpose(1, 2).contiguous().view(batch_size, -1, num_heads * dim_per_head)
        output = self.out(context)
        output = self.dropout(output) + residual
        output = self.layer_norm(output)
        return output, attention


class PositionWiseFeedForward(nn.Module):
    """
    Position-wise feed forward layer
    位置智能 前馈 层
    后置层标准化
    """

    def __init__(self, model_dim=512, ffn_dim=2048, dropout=0.0):
        super(PositionWiseFeedForward, self).__init__()
        self.w1 = nn.Conv1d(model_dim, ffn_dim, 1)
        self.w2 = nn.Conv1d(ffn_dim, model_dim, 1)
        self.dropout = nn.Dropout(dropout)
        self.layer_norm = nn.LayerNorm(model_dim)

    def forward(self, x):
        """
        :param x: [batch_size, seq_len, model_dim]
        return:
            [batch_size, seq_len, model_dim]
        """
        residual = x
        x = self.w1(x.transpose(1, 2))
        x = F.relu(x)
        x = self.w2(x)
        x = self.dropout(x.transpose(1, 2))
        x = x + residual
        x = self.layer_norm(x)
        return x


class PositionalEncoding(nn.Module):
    """
    Positional encoding
    位置 编码
    """

    def __init__(self, model_dim, max_len):
        super(PositionalEncoding, self).__init__()
        """
        2.0 * (j // 2) : 含义: 得到j的向下取最近偶数的值, 并将整数转为float数, 5->4,6->6, 9->8
        """
        position_encoding = np.array(
            [
                [pos / np.power(10000, 2.0 * (j // 2) / model_dim)
                 for j in range(model_dim)]
                for pos in range(max_len)
            ]
        )
        position_encoding[:, 0::2] = np.sin(position_encoding[:, 0::2])
        position_encoding[:, 1::2] = np.cos(position_encoding[:, 1::2])
        pad_row = torch.zeros([1, model_dim])
        position_encoding = torch.cat([pad_row, torch.from_numpy(position_encoding)])
        self.position_encoding = nn.Embedding(max_len + 1, model_dim)
        self.position_encoding.weight = nn.Parameter(position_encoding, requires_grad=False)

    def forward(self, input_len):
        """
        Args:
            input_len (batch_size, 1): input length
        """
        max_len = torch.max(input_len)
        tensor = torch.cuda.LongTensor if input_len.is_cuda else torch.LongTensor
        input_pos = tensor(
            [list(range(1, lens.item() + 1)) + [0] * (max_len - lens.item()) for lens in input_len]
        )
        return self.position_encoding(input_pos)



import inspect
import math

import torch
import torch.nn as nn
from gxl_ai_utils.utils import utils_model
import torch.nn.functional as F


class Sequential(torch.nn.ModuleDict):
    """A sequence of modules with potentially inferring shape on construction.

    If layers are passed with names, these can be referenced with dot notation.

    Arguments
    ---------
    input_shape : iterable
        A list or tuple of ints or None, representing the expected shape of an
        input tensor. None represents a variable-length dimension. If no
        ``input_shape`` is passed, no shape inference will be performed.
    *layers, **named_layers
        The inputs are treated as a list of layers to be
        applied in sequence. The output shape of each layer is used to
        infer the shape of the following layer. If a tuple is returned,
        only the shape of the first element is used to determine input
        shape of the next layer (e.g. RNN returns output, hidden).

    Example
    -------
    >>> inputs = torch.rand(10, 40, 50)
    >>> model = Sequential(input_shape=inputs.shape)
    >>> model.append(Linear, n_neurons=100, layer_name="layer1")
    >>> model.append(Linear, n_neurons=200, layer_name="layer2")
    >>> outputs = model(inputs)
    >>> outputs.shape
    torch.Size([10, 40, 200])
    >>> outputs = model.layer1(inputs)
    >>> outputs.shape
    torch.Size([10, 40, 100])
    """

    def __init__(self, *layers, input_shape=None, **named_layers):
        super().__init__()

        # Make sure either layers or input_shape is passed
        if not layers and input_shape is None and not named_layers:
            raise ValueError("Must pass either layers or input shape")

        # Keep track of what layers need "lengths" passed
        self.length_layers = []

        # Replace None dimensions with arbitrary value
        self.input_shape = input_shape
        if input_shape and None in input_shape:
            self.input_shape = list(input_shape)
            for i, dim in enumerate(self.input_shape):

                # To reduce size of dummy tensors, use 1 for batch dim
                if i == 0 and dim is None:
                    dim = 1

                # Use 64 as nice round arbitrary value, big enough that
                # halving this dimension a few times doesn't reach 1
                self.input_shape[i] = dim or 256

        # Append non-named layers
        for layer in layers:
            self.append(layer)

        # Append named layers
        for name, layer in named_layers.items():
            self.append(layer, layer_name=name)

    def append(self, layer, *args, layer_name=None, **kwargs):
        """Add a layer to the list of layers, inferring shape if necessary.

        Arguments
        ---------
        layer : A torch.nn.Module class or object
            If the layer is a class, it should accept an argument called
            ``input_shape`` which will be inferred and passed. If the layer
            is a module object, it is added as-is.
        layer_name : str
            The name of the layer, for reference. If the name is in use,
            ``_{count}`` will be appended.
        *args, **kwargs
            These are passed to the layer if it is constructed.
        """

        # Compute layer_name
        if layer_name is None:
            layer_name = str(len(self))
        elif layer_name in self:
            index = 0
            while f"{layer_name}_{index}" in self:
                index += 1
            layer_name = f"{layer_name}_{index}"

        # Check if it needs to be constructed with input shape
        if self.input_shape:
            argspec = inspect.getfullargspec(layer)
            if "input_shape" in argspec.args + argspec.kwonlyargs:
                input_shape = self.get_output_shape()
                layer = layer(*args, input_shape=input_shape, **kwargs)

        # Finally, append the layer.
        try:
            self.add_module(layer_name, layer)
        except TypeError:
            raise ValueError(
                "Must pass `input_shape` at initialization and use "
                "modules that take `input_shape` to infer shape when "
                "using `append()`."
            )

    def get_output_shape(self):
        """Returns expected shape of the output.

        Computed by passing dummy input constructed with the
        ``self.input_shape`` attribute.
        """
        with torch.no_grad():
            dummy_input = torch.zeros(self.input_shape)
            dummy_output = self(dummy_input)
        return dummy_output.shape

    def forward(self, x):
        """Applies layers in sequence, passing only the first element of tuples.

        Arguments
        ---------
        x : torch.Tensor
            The input tensor to run through the network.
        """
        for layer in self.values():
            x = layer(x)
            if isinstance(x, tuple):
                x = x[0]

        return x


class Softmax(torch.nn.Module):
    """Computes the softmax of a 2d, 3d, or 4d input tensor.

    Arguments
    ---------
    apply_log : bool
        Whether to apply the log function before softmax.
    dim : int
        If the dimension where softmax is applied.
    reshape: bool
        whether to apply reshaping (true by default)

    Example
    -------
    >>> classifier = Softmax()
    >>> inputs = torch.rand(10, 50, 40)
    >>> output = classifier(inputs)
    >>> output.shape
    torch.Size([10, 50, 40])
    """

    def __init__(self, apply_log=False, dim=-1, reshape=True):
        super().__init__()

        if apply_log:
            self.act = torch.nn.LogSoftmax(dim=dim)
        else:
            self.act = torch.nn.Softmax(dim=dim)

        self.reshape = reshape

    def forward(self, x):
        """Returns the softmax of the input tensor.

        Arguments
        ---------
        x : torch.Tensor
            Input tensor.
        """
        # Reshaping the tensors
        dims = x.shape

        if self.reshape:
            if len(dims) == 3:
                x = x.reshape(dims[0] * dims[1], dims[2])

            if len(dims) == 4:
                x = x.reshape(dims[0] * dims[1], dims[2], dims[3])

        x_act = self.act(x)

        # Retrieving the original shape format
        if self.reshape:
            if len(dims) == 3:
                x_act = x_act.reshape(dims[0], dims[1], dims[2])

            if len(dims) == 4:
                x_act = x_act.reshape(dims[0], dims[1], dims[2], dims[3])

        return x_act


def get_padding_elem(L_in: int, stride: int, kernel_size: int, dilation: int):
    """This function computes the number of elements to add for zero-padding.

    Arguments
    ---------
    L_in : int
    stride: int
    kernel_size : int
    dilation : int
    """
    if stride > 1:
        padding = [math.floor(kernel_size / 2), math.floor(kernel_size / 2)]

    else:
        L_out = (
                math.floor((L_in - dilation * (kernel_size - 1) - 1) / stride) + 1
        )
        padding = [
            math.floor((L_in - L_out) / 2),
            math.floor((L_in - L_out) / 2),
        ]
    return padding


class BatchNorm1d(nn.Module):
    """Applies 1d batch normalization to the input tensor.

    Arguments
    ---------
    input_shape : tuple
        The expected shape of the input. Alternatively, use ``input_size``.
    input_size : int
        The expected size of the input. Alternatively, use ``input_shape``.
    eps : float
        This value is added to std deviation estimation to improve the numerical
        stability.
    momentum : float
        It is a value used for the running_mean and running_var computation.
    affine : bool
        When set to True, the affine parameters are learned.
    track_running_stats : bool
        When set to True, this module tracks the running mean and variance,
        and when set to False, this module does not track such statistics.
    combine_batch_time : bool
        When true, it combines batch an time axis.


    Example
    -------
    >>> input = torch.randn(100, 10)
    >>> norm = BatchNorm1d(input_shape=input.shape)
    >>> output = norm(input)
    >>> output.shape
    torch.Size([100, 10])
    """

    def __init__(
            self,
            input_shape=None,
            input_size=None,
            eps=1e-05,
            momentum=0.1,
            affine=True,
            track_running_stats=True,
            combine_batch_time=False,
            skip_transpose=False,
    ):
        super().__init__()
        self.combine_batch_time = combine_batch_time
        self.skip_transpose = skip_transpose

        if input_size is None and skip_transpose:
            input_size = input_shape[1]
        elif input_size is None:
            input_size = input_shape[-1]

        self.norm = nn.BatchNorm1d(
            input_size,
            eps=eps,
            momentum=momentum,
            affine=affine,
            track_running_stats=track_running_stats,
        )

    def forward(self, x):
        """Returns the normalized input tensor.

        Arguments
        ---------
        x : torch.Tensor (batch, time, [channels])
            input to normalize. 2d or 3d tensors are expected in input
            4d tensors can be used when combine_dims=True.
        """
        shape_or = x.shape
        if self.combine_batch_time:
            if x.ndim == 3:
                x = x.reshape(shape_or[0] * shape_or[1], shape_or[2])
            else:
                x = x.reshape(
                    shape_or[0] * shape_or[1], shape_or[3], shape_or[2]
                )

        elif not self.skip_transpose:
            x = x.transpose(-1, 1)

        x_n = self.norm(x)

        if self.combine_batch_time:
            x_n = x_n.reshape(shape_or)
        elif not self.skip_transpose:
            x_n = x_n.transpose(1, -1)

        return x_n


class Conv1d(nn.Module):
    """This function implements 1d convolution.

    Arguments
    ---------
    out_channels : int
        It is the number of output channels.
    kernel_size : int
        Kernel size of the convolutional filters.
    input_shape : tuple
        The shape of the input. Alternatively use ``in_channels``.
    in_channels : int
        The number of input channels. Alternatively use ``input_shape``.
    stride : int
        Stride factor of the convolutional filters. When the stride factor > 1,
        a decimation in time is performed.
    dilation : int
        Dilation factor of the convolutional filters.
    padding : str
        (same, valid, causal). If "valid", no padding is performed.
        If "same" and stride is 1, output shape is the same as the input shape.
        "causal" results in causal (dilated) convolutions.
    groups: int
        Number of blocked connections from input channels to output channels.
    padding_mode : str
        This flag specifies the type of padding. See torch.nn documentation
        for more information.
    skip_transpose : bool
        If False, uses batch x time x channel convention of speechbrain.
        If True, uses batch x channel x time convention.
    weight_norm : bool
        If True, use weight normalization,
        to be removed with self.remove_weight_norm() at inference
    conv_init : str
        Weight initialization for the convolution network
    default_padding: str or int
        This sets the default padding mode that will be used by the pytorch Conv1d backend.

    Example
    -------
    >>> inp_tensor = torch.rand([10, 40, 16])
    >>> cnn_1d = Conv1d(
    ...     input_shape=inp_tensor.shape, out_channels=8, kernel_size=5
    ... )
    >>> out_tensor = cnn_1d(inp_tensor)
    >>> out_tensor.shape
    torch.Size([10, 40, 8])
    """

    def __init__(
            self,
            out_channels,
            kernel_size,
            input_shape=None,
            in_channels=None,
            stride=1,
            dilation=1,
            padding="same",
            groups=1,
            bias=True,
            padding_mode="reflect",
            skip_transpose=False,
            weight_norm=False,
            conv_init=None,
            default_padding=0,
    ):
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.padding = padding
        self.padding_mode = padding_mode
        self.unsqueeze = False
        self.skip_transpose = skip_transpose

        if input_shape is None and in_channels is None:
            raise ValueError("Must provide one of input_shape or in_channels")

        if in_channels is None:
            in_channels = self._check_input_shape(input_shape)

        self.in_channels = in_channels

        self.conv = nn.Conv1d(
            in_channels,
            out_channels,
            self.kernel_size,
            stride=self.stride,
            dilation=self.dilation,
            padding=default_padding,
            groups=groups,
            bias=bias,
        )

        if conv_init == "kaiming":
            nn.init.kaiming_normal_(self.conv.weight)
        elif conv_init == "zero":
            nn.init.zeros_(self.conv.weight)
        elif conv_init == "normal":
            nn.init.normal_(self.conv.weight, std=1e-6)

        if weight_norm:
            self.conv = nn.utils.weight_norm(self.conv)

    def forward(self, x):
        """Returns the output of the convolution.

        Arguments
        ---------
        x : torch.Tensor (batch, time, channel)
            input to convolve. 2d or 4d tensors are expected.
        """
        if not self.skip_transpose:
            x = x.transpose(1, -1)

        if self.unsqueeze:
            x = x.unsqueeze(1)

        if self.padding == "same":
            x = self._manage_padding(
                x, self.kernel_size, self.dilation, self.stride
            )

        elif self.padding == "causal":
            num_pad = (self.kernel_size - 1) * self.dilation
            x = F.pad(x, (num_pad, 0))

        elif self.padding == "valid":
            pass

        else:
            raise ValueError(
                "Padding must be 'same', 'valid' or 'causal'. Got "
                + self.padding
            )

        wx = self.conv(x)

        if self.unsqueeze:
            wx = wx.squeeze(1)

        if not self.skip_transpose:
            wx = wx.transpose(1, -1)

        return wx

    def _manage_padding(
            self, x, kernel_size: int, dilation: int, stride: int,
    ):
        """This function performs zero-padding on the time axis
        such that their lengths is unchanged after the convolution.

        Arguments
        ---------
        x : torch.Tensor
            Input tensor.
        kernel_size : int
            Size of kernel.
        dilation : int
            Dilation used.
        stride : int
            Stride.
        """

        # Detecting input shape
        L_in = self.in_channels

        # Time padding
        padding = get_padding_elem(L_in, stride, kernel_size, dilation)

        # Applying padding
        x = F.pad(x, padding, mode=self.padding_mode)

        return x

    def _check_input_shape(self, shape):
        """Checks the input shape and returns the number of input channels."""

        if len(shape) == 2:
            self.unsqueeze = True
            in_channels = 1
        elif self.skip_transpose:
            in_channels = shape[1]
        elif len(shape) == 3:
            in_channels = shape[2]
        else:
            raise ValueError(
                "conv1d expects 2d, 3d inputs. Got " + str(len(shape))
            )

        # Kernel size must be odd
        if not self.padding == "valid" and self.kernel_size % 2 == 0:
            raise ValueError(
                "The field kernel size must be an odd number. Got %s."
                % (self.kernel_size)
            )

        return in_channels

    def remove_weight_norm(self):
        """Removes weight normalization at inference if used during training."""
        self.conv = nn.utils.remove_weight_norm(self.conv)


class StatisticsPooling(nn.Module):
    """This class implements a statistic pooling layer.

    It returns the mean and/or std of input tensor.

    Arguments
    ---------
    return_mean : True
         If True, the average pooling will be returned.
    return_std : True
         If True, the standard deviation will be returned.

    Example
    -------
    >>> inp_tensor = torch.rand([5, 100, 50])
    >>> sp_layer = StatisticsPooling()
    >>> out_tensor = sp_layer(inp_tensor)
    >>> out_tensor.shape
    torch.Size([5, 1, 100])
    """

    def __init__(self, return_mean=True, return_std=True):
        super().__init__()

        # Small value for GaussNoise
        self.eps = 1e-5
        self.return_mean = return_mean
        self.return_std = return_std
        if not (self.return_mean or self.return_std):
            raise ValueError(
                "both of statistics are equal to False \n"
                "consider enabling mean and/or std statistic pooling"
            )

    def forward(self, x, lengths=None):
        """Calculates mean and std for a batch (input tensor).
        消除第2维度, 得到第二维度的mean和std, 也就是第三维度乘2
        Arguments
        ---------
        x : torch.Tensor
            It represents a tensor for a mini-batch.
        """
        if lengths is None:
            if self.return_mean:
                mean = x.mean(dim=1)
            if self.return_std:
                std = x.std(dim=1)
        else:
            mean = []
            std = []
            for snt_id in range(x.shape[0]):
                # Avoiding padded time steps
                actual_size = int(torch.round(lengths[snt_id] * x.shape[1]))

                # computing statistics
                if self.return_mean:
                    mean.append(
                        torch.mean(x[snt_id, 0:actual_size, ...], dim=0)
                    )
                if self.return_std:
                    std.append(torch.std(x[snt_id, 0:actual_size, ...], dim=0))
            if self.return_mean:
                mean = torch.stack(mean)
            if self.return_std:
                std = torch.stack(std)

        if self.return_mean:
            gnoise = self._get_gauss_noise(mean.size(), device=mean.device)
            gnoise = gnoise
            mean += gnoise
        if self.return_std:
            std = std + self.eps

        # Append mean and std of the batch
        if self.return_mean and self.return_std:
            pooled_stats = torch.cat((mean, std), dim=1)
            pooled_stats = pooled_stats.unsqueeze(1)
        elif self.return_mean:
            pooled_stats = mean.unsqueeze(1)
        elif self.return_std:
            pooled_stats = std.unsqueeze(1)

        return pooled_stats

    def _get_gauss_noise(self, shape_of_tensor, device="cpu"):
        """Returns a tensor of epsilon Gaussian noise.

        Arguments
        ---------
        shape_of_tensor : tensor
            It represents the size of tensor for generating Gaussian noise.
        """
        gnoise = torch.randn(shape_of_tensor, device=device)
        gnoise -= torch.min(gnoise)
        gnoise /= torch.max(gnoise)
        gnoise = self.eps * ((1 - 9) * gnoise + 9)

        return gnoise


class Linear(torch.nn.Module):
    """Computes a linear transformation y = wx + b.

    Arguments
    ---------
    n_neurons : int
        It is the number of output neurons (i.e, the dimensionality of the
        output).
    input_shape: tuple
        It is the shape of the input tensor.
    input_size: int
        Size of the input tensor.
    bias : bool
        If True, the additive bias b is adopted.
    combine_dims : bool
        If True and the input is 4D, combine 3rd and 4th dimensions of input.
    max_norm: float
        weight max-norm.
    Example
    -------
    >>> inputs = torch.rand(10, 50, 40)
    >>> lin_t = Linear(input_shape=(10, 50, 40), n_neurons=100)
    >>> output = lin_t(inputs)
    >>> output.shape
    torch.Size([10, 50, 100])
    """

    def __init__(
            self,
            n_neurons,
            input_shape=None,
            input_size=None,
            bias=True,
            max_norm=None,
            combine_dims=False,
    ):
        super().__init__()
        self.max_norm = max_norm
        self.combine_dims = combine_dims

        if input_shape is None and input_size is None:
            raise ValueError("Expected one of input_shape or input_size")

        if input_size is None:
            input_size = input_shape[-1]
            if len(input_shape) == 4 and self.combine_dims:
                input_size = input_shape[2] * input_shape[3]

        # Weights are initialized following pytorch approach
        self.w = nn.Linear(input_size, n_neurons, bias=bias)

    def forward(self, x):
        """Returns the linear transformation of input tensor.

        Arguments
        ---------
        x : torch.Tensor
            Input to transform linearly.
        """
        if x.ndim == 4 and self.combine_dims:
            x = x.reshape(x.shape[0], x.shape[1], x.shape[2] * x.shape[3])

        if self.max_norm is not None:
            self.w.weight.data = torch.renorm(
                self.w.weight.data, p=2, dim=0, maxnorm=self.max_norm
            )

        wx = self.w(x)

        return wx


class Xvector(torch.nn.Module):
    """This model extracts X-vectors for speaker recognition and diarization.

    Arguments
    ---------
    device : str
        Device used e.g. "cpu" or "cuda".
    activation : torch class
        A class for constructing the activation layers.
    tdnn_blocks : int
        Number of time-delay neural (TDNN) layers.
    tdnn_channels : list of ints
        Output channels for TDNN layer.
    tdnn_kernel_sizes : list of ints
        List of kernel sizes for each TDNN layer.
    tdnn_dilations : list of ints
        List of dilations for kernels in each TDNN layer.
    lin_neurons : int
        Number of neurons in linear layers.

    Example
    -------
    >>> compute_xvect = Xvector('cpu')
    >>> input_feats = torch.rand([5, 10, 40])
    >>> outputs = compute_xvect(input_feats)
    >>> outputs.shape
    torch.Size([5, 1, 512])
    """

    def __init__(
            self,
            activation=torch.nn.LeakyReLU,
            tdnn_blocks=5,
            tdnn_channels=[512, 512, 512, 512, 1500],
            tdnn_kernel_sizes=[5, 3, 3, 1, 1],
            tdnn_dilations=[1, 2, 3, 1, 1],
            lin_neurons=512,
            in_channels=80,
    ):

        super().__init__()
        self.blocks = nn.ModuleList()

        # TDNN layers
        for block_index in range(tdnn_blocks):
            out_channels = tdnn_channels[block_index]
            self.blocks.extend(
                [
                    Conv1d(
                        in_channels=in_channels,
                        out_channels=out_channels,
                        kernel_size=tdnn_kernel_sizes[block_index],
                        dilation=tdnn_dilations[block_index],
                    ),
                    activation(),
                    BatchNorm1d(input_size=out_channels),
                ]
            )
            in_channels = tdnn_channels[block_index]
        # tdnn layers -14个time
        # Statistical pooling
        self.blocks.append(StatisticsPooling())

        # Final linear transformation
        self.blocks.append(
            Linear(
                input_size=out_channels * 2,
                n_neurons=lin_neurons,
                bias=True,
                combine_dims=False,
            )
        )

    def forward(self, x, lens=None):
        """Returns the x-vectors.

        Arguments
        ---------
        x : torch.Tensor
        """

        for layer in self.blocks:
            try:
                x = layer(x, lengths=lens)
            except TypeError:
                x = layer(x)
        return x


class Classifier(Sequential):
    """This class implements the last MLP on the top of xvector features.

    Arguments
    ---------
    input_shape : tuple
        Expected shape of an example input.
    activation : torch class
        A class for constructing the activation layers.
    lin_blocks : int
        Number of linear layers.
    lin_neurons : int
        Number of neurons in linear layers.
    out_neurons : int
        Number of output neurons.

    Example
    -------
    >>> input_feats = torch.rand([5, 10, 40])
    >>> compute_xvect = Xvector()
    >>> xvects = compute_xvect(input_feats)
    >>> classify = Classifier(input_shape=xvects.shape)
    >>> output = classify(xvects)
    >>> output.shape
    torch.Size([5, 1, 1211])
    """

    def __init__(
            self,
            input_shape,
            activation=torch.nn.LeakyReLU,
            lin_blocks=1,
            lin_neurons=512,
            out_neurons=2,
    ):
        super().__init__(input_shape=input_shape)

        self.append(activation(), layer_name="act")
        self.append(BatchNorm1d, layer_name="norm")

        if lin_blocks > 0:
            self.append(Sequential, layer_name="DNN")

        for block_index in range(lin_blocks):
            block_name = f"block_{block_index}"
            self.DNN.append(
                Sequential, layer_name=block_name
            )
            self.DNN[block_name].append(
                Linear,
                n_neurons=lin_neurons,
                bias=True,
                layer_name="linear",
            )
            self.DNN[block_name].append(activation(), layer_name="act")
            self.DNN[block_name].append(
                BatchNorm1d, layer_name="norm"
            )

        # Final Softmax classifier
        self.append(
            Linear, n_neurons=out_neurons, layer_name="out"
        )
        self.append(
            Softmax(apply_log=True), layer_name="softmax"
        )


class GxlClassifier(nn.Module):
    def __init__(self):
        super().__init__()
        self.emb = Xvector()
        self.classifier = Classifier((None, None, 512))

    def forward(self, x):
        embedded = self.emb(x)
        return self.classifier(embedded).squeeze(1)


if __name__ == "__main__":
    """
    512
    100-> 96->92->86
    1
    172
    """
    inputs = torch.randn(3, 100, 80)
    model = Xvector()
    outputs = model(inputs)
    print(outputs.shape)
    classifier = Classifier((None, None, 512))
    output2 = classifier(outputs)
    print(output2.shape)
    model_final = GxlClassifier()
    output3 = model_final(inputs)
    print(output3.shape)
    # model = StatisticsPooling()
    # inputs2 = torch.randn(3, 1500, 11200)
    # outputs2 = model(inputs2)
    # print(outputs2.shape)
    # inputs2 = torch.randn(3, 100, 1101)
    # outputs2 = model(inputs2)
    # print(outputs2.shape)



from typing import Tuple

import torch
from torch import nn
from .embedding import PositionalEncoding
from gxl_ai_utils.utils import utils_model


class BaseSubsampling(nn.Module):
    def __init__(self, pos_enc_class: PositionalEncoding):
        super().__init__()
        self.pos_enc = pos_enc_class
        # window_size = (1 + right_context) + (chunk_size -1) * subsampling_rate
        self.right_context = 0
        # stride = subsampling_rate * chunk_size
        self.subsampling_rate = 1

    def position_encoding(self, offset: int, size: int) -> torch.Tensor:
        return self.pos_enc.position_encoding(offset, size)


class Conv2dSubsampling(BaseSubsampling):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Conv2dSubsampling4(Conv2dSubsampling):
    """Convolutional 2D subsampling (to 1/4 length)."""

    def __init__(self,
                 idim: int,
                 odim: int,
                 dropout_rate: float,
                 pos_enc_class: PositionalEncoding = None):
        """Construct an Conv2dSubsampling4 object.

        Args:
            idim (int): Input dimension.
            odim (int): Output dimension.
            dropout_rate (float): Dropout rate.
        """
        if pos_enc_class is None:
            pos_enc_class = PositionalEncoding(odim, dropout_rate)
        super().__init__(pos_enc_class)
        self.conv = nn.Sequential(
            nn.Conv2d(1, odim, kernel_size=3, stride=2),
            nn.ReLU(),
            nn.Conv2d(odim, odim, kernel_size=3, stride=2),
            nn.ReLU(), )
        self.out = nn.Sequential(
            #  计算公式: (input - filter + 2 * padding)/stride + 1
            nn.Linear(odim * (((idim - 1) // 2 - 1) // 2), odim))
        self.subsampling_rate = 4
        # The right context for every conv layer is computed by:
        # (kernel_size - 1) * frame_rate_of_this_layer
        # 6 = (3 - 1) * 1 + (3 - 1) * 2
        self.right_context = 6

    def forward(self, x: torch.Tensor, x_mask: torch.Tensor, offset: int = 0
                ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Subsample x.
        Args:
            x (torch.Tensor): Input tensor (#batch, time, idim).
            x_mask (torch.Tensor): Input mask (#batch, 1, time).
            offset (int): position encoding offset.
        Returns:
            torch.Tensor: Subsampled tensor (#batch, time', odim),
                where time' = time // 4.
            torch.Tensor: positional encoding
            torch.Tensor: Subsampled mask (#batch, 1, time'),
                where time' = time // 4.
        """
        x = x.unsqueeze(1)  # (b, c=1, t, f)
        x = self.conv(x)
        b, c, t, f = x.shape
        x = self.out(x.transpose(2, 1).reshape(b, -1, c * f))
        x, pos_emb = self.pos_enc(x, offset)
        return x, pos_emb, x_mask[:, :, :-2:2][:, :, :-2:2]


class Conv2dSubsampling4Pure(Conv2dSubsampling4):
    """
    纯粹的conv2d subsampling, 不考虑位置编码和mask
    """

    def __init__(self, idim, odim, dropout_rate):
        super().__init__(idim, odim, dropout_rate, None)
        self.output_dim = utils_model.get_output_dim_for_conv(utils_model.get_output_dim_for_conv(idim, 3, 2), 3,
                                                              2) * odim
        self.receptive_field_length = utils_model.get_receptive_field_size_for_continuous_conv([3, 3], [2, 2])

    def forward(self, x: torch.Tensor, x_len: torch.Tensor
                ) -> Tuple[torch.Tensor, torch.Tensor]:
        x = x.unsqueeze(1)  # (b, c=1, t, f)
        x = self.conv(x)
        b, c, t, f = x.shape
        x = x.transpose(2, 1).reshape(b, -1, c * f)  # (b, t, f*c)
        x_len = utils_model.get_output_dim_for_conv(utils_model.get_output_dim_for_conv(x_len, 3, 2), 3,
                                                    2)
        return x, x_len


class Conv2dSubsampling6(Conv2dSubsampling):
    """Convolutional 2D subsampling (to 1/6 length)."""

    def __init__(self,
                 idim: int,
                 odim: int,
                 dropout_rate: float,
                 pos_enc_class: PositionalEncoding = None):
        """Construct an Conv2dSubsampling6 object.

        Args:
            idim (int): Input dimension.
            odim (int): Output dimension.
            dropout_rate (float): Dropout rate.
        """
        if pos_enc_class is None:
            pos_enc_class = PositionalEncoding(odim, dropout_rate)
        super().__init__(pos_enc_class)
        self.conv = nn.Sequential(
            nn.Conv2d(1, odim, kernel_size=3, stride=2),
            nn.ReLU(),
            nn.Conv2d(odim, odim, kernel_size=5, stride=3),
            nn.ReLU(), )
        conv_output_dim = utils_model.get_output_dim_for_conv(utils_model.get_output_dim_for_conv(idim, 3, 2), 5, 3)
        self.out = nn.Sequential(
            #  计算公式: (input - filter + 2 * padding)/stride + 1
            nn.Linear(odim * conv_output_dim, odim))
        # The right context for every conv layer is computed by:
        # (kernel_size - 1) * frame_rate_of_this_layer
        # 10 = (3 - 1) * 1 + (5 - 1) * 2
        self.subsampling_rate = 6
        self.right_context = 10

    def forward(self, x: torch.Tensor, x_mask: torch.Tensor, offset: int = 0
                ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Subsample x.
        Args:
            x (paddle.Tensor): Input tensor (#batch, time, idim).
            x_mask (paddle.Tensor): Input mask (#batch, 1, time).
            offset (int): position encoding offset.
        Returns:
            Tensor: Subsampled tensor (#batch, time', odim),
                where time' = time // 6.
            Tensor: positional encoding
            Tensor: Subsampled mask (#batch, 1, time'),
                where time' = time // 6.
        """
        x = x.unsqueeze(1)  # (b, c=1, t, f)
        x = self.conv(x)
        b, c, t, f = x.shape
        x = self.out(x.transpose(2, 1).reshape(b, -1, c * f))
        x, pos_emb = self.pos_enc(x, offset)
        #  关于x_mask的取值: -2 = filter1-1, -4 = filter2-1, 2 = stride1, 3=stride2
        return x, pos_emb, x_mask[:, :, :-2:2][:, :, :-4:3]


class Conv2dSubsampling8(Conv2dSubsampling):
    """Convolutional 2D subsampling (to 1/8 length)."""

    def __init__(self,
                 idim: int,
                 odim: int,
                 dropout_rate: float,
                 pos_enc_class: PositionalEncoding = None):
        """Construct an Conv2dSubsampling8 object.

        Args:
            idim (int): Input dimension.
            odim (int): Output dimension.
            dropout_rate (float): Dropout rate.
        """
        if pos_enc_class is None:
            pos_enc_class = PositionalEncoding(odim, dropout_rate)
        super().__init__(pos_enc_class)
        self.conv = nn.Sequential(
            nn.Conv2d(1, odim, kernel_size=3, stride=2),
            nn.ReLU(),
            nn.Conv2d(odim, odim, kernel_size=3, stride=2),
            nn.ReLU(),
            nn.Conv2d(odim, odim, kernel_size=3, stride=2),
            nn.ReLU(),
        )
        conv_output_dim = utils_model.get_output_dim_for_conv(
            utils_model.get_output_dim_for_conv(utils_model.get_output_dim_for_conv(idim, 3, 2), 3, 2), 3, 2)
        self.out = nn.Sequential(
            #  计算公式: (input - filter + 2 * padding)/stride + 1
            nn.Linear(odim * conv_output_dim, odim))
        # The right context for every conv layer is computed by:
        # (kernel_size - 1) * frame_rate_of_this_layer
        # 14 = (3 - 1) * 1 + (3 - 1) * 2 + (3 - 1) * 4 = 2 + 4 + 8
        self.subsampling_rate = 8
        self.right_context = 14

    def forward(self, x: torch.Tensor, x_mask: torch.Tensor, offset: int = 0
                ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Subsample x.
        Args:
            x (paddle.Tensor): Input tensor (#batch, time, idim).
            x_mask (paddle.Tensor): Input mask (#batch, 1, time).
            offset (int): position encoding offset.
        Returns:
            Tensor: Subsampled tensor (#batch, time', odim),
                where time' = time // 8.
            Tensor: positional encoding
            Tensor: Subsampled mask (#batch, 1, time'),
                where time' = time // 8.
        """
        x = x.unsqueeze(1)  # (b, c=1, t, f)
        x = self.conv(x)
        b, c, t, f = x.shape
        x = self.out(x.transpose(2, 1).reshape(b, -1, c * f))
        x, pos_emb = self.pos_enc(x, offset)
        return x, pos_emb, x_mask[:, :, :-2:2][:, :, :-2:2][:, :, :-2:2]



import os
import threading
import time
import traceback

from sortedcontainers import SortedSet

from gxl_ai_utils.utils import utils_spider, utils_file
from gxl_ai_utils.AiConstant import AI_LOGGER

logger = AI_LOGGER()

header = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "zh-CN,zh;q=0.9,en;q=0.8",
    "Cache-Control": "max-age=0",
    "Connection": "keep-alive",
    # "Cookie": "_xmLog=h5&bbc1105d-db8f-4664-96af-86d46a1abadc&process.env.sdkVersion; 1&remember_me=y; 1&_token=293413321&4BAEDC40240N498E25AC5DD3EFA50F93A7308CB5D7574461E394BF9F4EB33C364865E35BA60C201MB4CD649CD4B91E3_; x-ats=ACNlNGE4Njc2MjAzZDdlM2Q1BTaNX3LXtAl4bXdlYl93d3c; xm-page-viewid=ximalaya-web; impl=www.ximalaya.com.login; x_xmly_traffic=utm_source%253A%2526utm_medium%253A%2526utm_campaign%253A%2526utm_content%253A%2526utm_term%253A%2526utm_from%253A; Hm_lvt_4a7d8ec50cfd6af753c4f8aee3425070=1696256199; Hm_lpvt_4a7d8ec50cfd6af753c4f8aee3425070=1696256199; web_login=1696256314748",
    "Host": "www.ximalaya.com",
    "Sec-Fetch-Dest": "document",
    "Sec-Fetch-Mode": "navigate",
    "Sec-Fetch-Site": "same-origin",
    "Sec-Fetch-User": "?1",
    "Upgrade-Insecure-Requests": "1",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-platform": "Windows",
}
#  下载wav时用的头
header2 = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "zh-CN,zh;q=0.9,en;q=0.8",
    "Cache-Control": "max-age=0",
    "If-None-Match": "75a67e32a04595ac0c393b4c5ba53064",
    "Range": "bytes=0-104857500",  # 默认每个音频小于100兆, 多出的部分直接截断
    "Sec-Ch-Ua-Mobile": "?0",
    "Sec-Ch-Ua-Platform": "Windows",
    "Sec-Fetch-Dest": "document",
    "Sec-Fetch-Mode": "navigate",
    "Sec-Fetch-Site": "none",
    "Sec-Fetch-User": "?1",
    "Upgrade-Insecure-Requests": "1",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/117.0.0.0 Safari/537.36",

}

SAVE_PATH_OLD = utils_file.join_path("/home/work_nfs6/xlgeng/data", "pachong_ximalaya_2", "wav_files")
SAVE_PATH_OLD2 = '/home/backup_nfs5/xlgeng/asr_data/ximalaya'
SAVE_PATH = utils_file.join_path("/home/node36_data", "ximalaya")
# SAVE_PATH = join_path_plus(".", "output")
utils_file.makedir_for_file_or_dir(SAVE_PATH)
utils_file.makedir_for_file_or_dir(SAVE_PATH_OLD)
PAGE_SIZE = 30
SORT_TYPE = 0  # 0: 正序,1:倒序
SAVE_WITH_TITLE = True
SAVE_WITH_ALBUM_ID = True


def get_name_by_rules(id, album_id, title, with_album_id, with_title):
    adds = []
    if with_album_id:
        adds.append(album_id)
    if with_title:
        title = utils_file.get_clean_filename(title)
        adds.append(title)
    name = str(id)
    for item in adds:
        name = name + '_' + item
    name = name + '.wav'
    return name


def if_response_content_is_none(response):
    if response is None or response.content is None:
        logger.info('response or response.content为None,加载数据失败')
        return True
    return False


def if_json_data_is_none(json_data):
    if json_data is None:
        logger.info('json_data is None, 未获取到json数据')
        return True
    return False


def get_ids(datas):
    """
    datas: [strs]:str_list or str:str
    return : [ids]:list
    """
    if isinstance(datas, list):
        return [x.split('/')[-1] for x in datas]
    elif isinstance(datas, str):
        return [datas.split('/')[-1]]


def save_audio_from_url(category, src, id, album_id, title=""):
    audio_dir = utils_file.join_path(SAVE_PATH, category)
    utils_file.makedir_sil(audio_dir)
    # src = "https://aod.cos.tx.xmcdn.com/group30/M0B/8E/06/wKgJXlmENsjB7SPHAP-6990Rfew647-aacv2-48K.m4a"
    file_name = get_name_by_rules(id, album_id, title, SAVE_WITH_ALBUM_ID, SAVE_WITH_TITLE)
    logger.info(f'开始保存:file_name:{file_name},save_dir:{audio_dir}, album_id:{album_id},src:{src}')
    start = time.time()
    file_path = utils_file.join_path(audio_dir, file_name)
    file_path_old = utils_file.join_path(SAVE_PATH_OLD, category, file_name)
    file_path_old2 = utils_file.join_path(SAVE_PATH_OLD2, category, file_name)
    if os.path.exists(file_path) or os.path.exists(file_path_old) or os.path.exists(file_path_old2):
        logger.info(f'音频文{file_name}已存在, 不在需要请求该音频.')
        return
    response_audio = utils_spider.send_request(src, headers=header2)
    if response_audio is None:
        return
    if if_response_content_is_none(response_audio):
        return
    with open(file_path, 'wb') as f:
        f.write(response_audio.content)
    end = time.time()
    logger.info(f'结束保存, 用时:{end - start}秒,url:{src}, file_name:{file_name}')
    # logger.info("休眠1秒钟, 防止被封")
    logger.info("不再休眠,继续干..")
    # time.sleep(1)


def get_wav_from_ids(category, ids, album_id, titles=None):
    if titles is None:
        titles = []
    logger.info(f'开始利用异步请求从ids得到每个id对应的音频路径')
    if len(ids) != len(titles):
        logger.info("ids与titles长度不一致, title并不重要, 使其强行一致即可")
        if len(ids) > len(titles):
            for _ in range(len(ids) - len(titles)):
                titles.append("")
        else:
            titles = titles[:len(ids)]
    for id, title in zip(ids, titles):
        logger.info(f'开始加载音频{id}的json数据')
        url = f"https://www.ximalaya.com/revision/play/v1/audio?id={id}&ptype=1"
        response = utils_spider.send_request(url, headers=header)
        if response is None:
            continue

        json_res = response.json()
        if if_json_data_is_none(json_res):
            continue
        else:
            logger.info(f'音频{id}的json数据加载成功')

        try:
            src = json_res['data']['src']
        except Exception as e:
            logger.info(f'音频{id}的json数据解析失败,未在指定路径得到音频地址')
            continue
        logger.info(f'音频{id}的json数据解析成功,开始保存音频')
        try:
            save_audio_from_url(category, src, id, album_id, title)
        except Exception as e:
            logger.info(f'音频{id}的保存函数save_audio_from_url执行报错, 爆错内容为{e}')
            traceback.print_exc()


def handle_alum_dependent_static_page(category, url):
    """
    通过解析静态页面获得音频地址,有错误的风险，获取的html文件并不一定是在浏览器看到的样子，可能是没渲染的状态
    url: 'https://www.ximalaya.com/album/9741525'
    """
    response = utils_spider.send_request(url, headers=header)
    if if_response_content_is_none(response):
        return
    logger.info('加载成功, 开始处理专辑')
    text = response.text
    tree = utils_spider.text2special_file(text)
    if tree is None:
        return
    xpath = "//li[@class='b_t']/div[@class='text b_t']/a/@to"
    datas = tree.xpath(xpath)
    ids = get_ids(datas)
    logger.info(f'处理专辑成功,得到该专辑所有音频id为:{ids}')
    get_wav_from_ids(category, ids, get_ids(url)[-1])


def handle_alum(category, album_id: str | int):
    """
    根据json数据获取信息
    album_id:型如 '9723091'
    分页获取数据
    """
    url_data = f"https://www.ximalaya.com/revision/album/v1/getTracksList?albumId={str(album_id)}&pageNum={1}&sort={SORT_TYPE}&pageSize={PAGE_SIZE}"
    logger.info(f'开始处理专辑(伴随分页模式),album_id:{album_id} ========')
    logger.info('开始获取总页数')
    response = utils_spider.send_request(url_data, headers=header)
    if if_response_content_is_none(response):
        return
    json_data = response.json()
    if if_json_data_is_none(json_data):
        return
    try:
        total_num = json_data['data']['trackTotalCount']
    except Exception as e:
        logger.info('json_data解析失败,未在指定路径获取该专辑音频总数')
        return
    total_page = int(total_num / 30) + 1
    logger.info(f'总页数为:{total_page}')
    for i in range(1, total_page + 1):
        logger.info(f'开始处理专辑{album_id}的第{i}页')
        url = f"https://www.ximalaya.com/revision/album/v1/getTracksList?albumId={album_id}&pageNum={i}&sort={SORT_TYPE}&pageSize={PAGE_SIZE}"
        response = utils_spider.send_request(url, headers=header)
        if if_response_content_is_none(response):
            continue
        json_data = response.json()
        if if_json_data_is_none(json_data):
            continue
        try:
            data_list = json_data['data']['tracks']
            ids_list = [track['trackId'] for track in data_list]
            title_list = [track['title'] for track in data_list]
        except Exception as e:
            logger.info('json_data解析失败,未在指定路径获取该专辑音频id列表')
            continue
        ids = ids_list
        logger.info(f'处理专辑成功,得到该专辑所有音频id为:{ids}')
        logger.info(f'处理专辑成功,得到该专辑所有音频titles为:{title_list}')
        try:
            get_wav_from_ids(category, ids, album_id, title_list)
        except Exception as e:
            logger.info(f'专辑{album_id}的get_wav_from_ids函数执行报错, 爆错内容为{e}')
            traceback.print_exc()


def handle_category_index(category, url):
    logger.info(f'开始加载 {category} 主页:' + url + ' ------------------------------')
    response = utils_spider.send_request(url, headers=header)
    if if_response_content_is_none(response):
        return
    logger.info(f'加载成功, 开始处理 {category} 主页')
    response.encoding = 'utf-8'
    text = response.text  # 相声主页
    tree = utils_spider.text2special_file(text)
    if tree is None:
        return
    xpath = '//ul[@class="_ZV"]/li[@class="_ZV"]/div[@class="album-wrapper   undefined T_G"]//div[@class="album-wrapper-card T_G"]/a[@class="album-cover   lg needhover _hW"]'
    datas = tree.xpath(xpath)
    xpath_for_title = '//*[@id="award"]/main/div[1]/div[3]/div[2]/ul/li/div/a[1]/span/text()'
    titles = tree.xpath(xpath_for_title)
    res = []
    res2 = []
    res3 = []
    titles_res = []
    for data in datas:
        res.append(data.xpath(".//@href"))
        res2.append(data.xpath(".//img[@class='corner-img _hW']/@src"))  # 付费专辑特有图标
        for re1, re2, title in zip(res, res2, titles):
            if len(re2) == 0:
                res3.append(re1)
                titles_res.append(title)
    res3 = [x[0] for x in res3]
    res3 = SortedSet(res3)
    logger.info(f"处理{category}主页成功, 得到当前分页的每个免费专辑的地址:" + str(res3))  # 删除付费专辑
    # logger.info("处理主页成功, 得到当前分页的每个免费专辑的地址:" + str(titles))  # 删除付费专辑
    for data in res3:
        album_id = -1
        try:
            album_id = get_ids(data)[-1]
            handle_alum(category, album_id)
        except Exception as e:
            logger.info(f'主页:{url}的专辑{album_id}的handle_alum函数执行报错, 爆错内容为{e}')
            traceback.print_exc()


def handle_category_index_fenye(category, url):
    """
    对类别主页进行分页处理
    """
    xpath = '//*[@id="award"]/main/div[1]/div[3]/div[3]/nav/div/form/input[2]/@max'
    response = utils_spider.send_request(url, headers=header)
    res = utils_spider.handle_xpath(response.text, xpath)
    num_fenye = int(res[0])
    try:
        handle_category_index(category, url)
    except Exception as e:
        logger.info(f'分页1的handle_category_index函数执行报错, 爆错内容为{e}')
        traceback.print_exc()
    for i in range(2, num_fenye + 1):
        try:
            handle_category_index(category, utils_file.join_path(url, f'p{i}'))
        except Exception as e:
            logger.info(f'分页{i}的handle_category_index函数执行报错, 爆错内容为{e}')
            traceback.print_exc()


def handle_big_index(url=None):
    url = "https://www.ximalaya.com/category"
    xpath_href = '//*[@id="award"]/main/div[1]/div[2]/div/a/@href'
    xpath_title = '//*[@id="award"]/main/div[1]/div[2]/div/a/text()'
    response = utils_spider.send_request(url, headers=header)
    tree = utils_spider.text2special_file(response.text)
    hrefs = tree.xpath(xpath_href)
    titles = tree.xpath(xpath_title)
    datas = {}
    for title, href in zip(titles, hrefs):
        datas[title] = utils_file.join_path(url, href.split('/')[2])
    print(datas)
    for item in datas.items():
        if item[0] == "音乐":
            logger.info('不再处理音乐专辑音乐')
            continue
        handle_category_index_fenye(item[0], item[1])


def handle_big_index_with_multi_thread(url=None):
    url = "https://www.ximalaya.com/category"
    xpath_href = '//*[@id="award"]/main/div[1]/div[2]/div/a/@href'
    xpath_title = '//*[@id="award"]/main/div[1]/div[2]/div/a/text()'
    response = utils_spider.send_request(url, headers=header)
    tree = utils_spider.text2special_file(response.text)
    hrefs = tree.xpath(xpath_href)
    titles = tree.xpath(xpath_title)
    datas = {}
    for title, href in zip(titles, hrefs):
        datas[title] = utils_file.join_path(url, href.split('/')[2])
    print(datas)
    threads = []
    for item in datas.items():
        if item[0] == "音乐":
            logger.info('不再处理音乐专辑音乐')
            continue
        thread = threading.Thread(target=handle_category_index_fenye, args=(item[0], item[1]))
        thread.start()
        threads.append(thread)
    for thread in threads:
        thread.join()



import os
import random
import sys
sys.path.append("/home/work_nfs7/xlgeng/code_runner_gxl/gxl_ai_utils/")

import torch
from gxl_ai_utils.gxl_model_wenet.transformer.encoder import TransformerEncoder
from transformers import AutoModelForCausalLM, AutoTokenizer

from gxl_ai_utils.utils import utils_file, utils_data
from gxl_ai_utils.thread.my_thread import GxlDynamicThreadPool


def do_prepare_data():
    aishell4_raw_scp_dir = '/home/work_nfs7/xlgeng/workspace/wenet_whisper/examples/aishell/s0/dump3/raw/aishell4'
    aishell4_raw_now = '/home/work_nfs6/xlgeng/data/data_scp/aishell4'
    utils_file.makedir_sil(aishell4_raw_now)
    aishell4_prefix = "/home/work_nfs7/xlgeng/workspace/wenet_whisper/examples/aishell/s0/"
    wav_scp = utils_file.load_dict_from_scp(utils_file.join_path(aishell4_raw_scp_dir, "wav.scp"))
    for k, v in wav_scp.items():
        wav_scp[k] = utils_file.join_path(aishell4_prefix, v)
    utils_file.write_dict_to_scp(wav_scp, utils_file.join_path(aishell4_raw_now, "wav.scp"))
    utils_file.copy_file(utils_file.join_path(aishell4_raw_scp_dir, "text"),
                         utils_file.join_path(aishell4_raw_now, "text"))

    ali_near_dir = "/home/work_nfs7/xlgeng/workspace/wenet_whisper/examples/aishell/s0/dump4/raw/Train_Ali_near"
    ali_far_dir = "/home/work_nfs7/xlgeng/workspace/wenet_whisper/examples/aishell/s0/dump2/raw/Train_Ali_far"
    # ali_far_dir_gxl = "/home/work_nfs6/xlgeng/data/data_scp/Train_Ali_far"
    # ali_near_dir_gxl = "/home/work_nfs6/xlgeng/data/data_scp/Train_Ali_near"
    # for file_name in ['wav.scp', 'text']:
    #     utils_file.copy_file(os.path.join(ali_far_dir, file_name), os.path.join(ali_far_dir_gxl, file))
    #     utils_file.copy_file(os.path.join(ali_near_dir, file_name), os.path.join(ali_near_dir_gxl, file))
    #
    # aishell4_dir_gxl = "/home/work_nfs6/xlgeng/data/data_scp/aishell4"

    runer = GxlDynamicThreadPool()
    output_root_dir = '/home/41_data/xlgeng/data/shards'
    utils_file.makedir_sil(output_root_dir)
    for dataset_dir in [aishell4_raw_now, ali_far_dir, ali_near_dir]:
        dataset_name = os.path.basename(dataset_dir)
        utils_file.logging_print(f"dataset_name: {dataset_name}")
        wav_scp_path = os.path.join(dataset_dir, 'wav.scp')
        text_path = os.path.join(dataset_dir, 'text')
        output_dir = os.path.join(output_root_dir, dataset_name)
        runer.add_task(utils_data.do_make_shard_file, [wav_scp_path, text_path, output_dir, 1000])
    runer.start()


def do_concat_shards_list():
    now_shard_list__path = "/home/work_nfs7/xlgeng/bsmu_template/wenet_SALMONN/examples/librispeech/salmonn/data/train/data.list.all.shuf"
    now_shard_list__path_2 = "/home/work_nfs7/xlgeng/bsmu_template/wenet_SALMONN/examples/librispeech/salmonn/data/train/data_2.list.all.shuf"
    data_dir = "/home/41_data/xlgeng/data/shards/"
    now_list = utils_file.load_list_file_clean(now_shard_list__path)
    dir_name_list = os.listdir(data_dir)
    for dir_name in dir_name_list:
        list_path = os.path.join(data_dir, dir_name, "shards_list.txt")
        temp_list = utils_file.load_list_file_clean(list_path)
        now_list.extend(temp_list)
    random.shuffle(now_list)
    utils_file.write_list_to_file(now_list, now_shard_list__path_2)


def do_test_model():
    import os

    tokenizer = AutoTokenizer.from_pretrained("/home/local_data/Atom-7B")
    model = AutoModelForCausalLM.from_pretrained("/home/local_data/Atom-7B")
    print(model)
    utils_file.print_model_size(model)


def do_test_model_2():
    speech_transformer = TransformerEncoder(
        input_size=1280,
        output_size=1280,
        attention_heads=4,
        linear_units=2560,
        num_blocks=4,
        dropout_rate=0.1,
        positional_dropout_rate=0.1,
        attention_dropout_rate=0.0,
        input_layer="linear",
        pos_enc_layer_type="abs_pos",
        normalize_before=True
    )
    utils_file.print_model_size(speech_transformer)


if __name__ == '__main__':
    """"""
    # do_prepare_data()
    # do_concat_shards_list()
    # do_test_model_2()
    do_test_model()


# Copyright (c) 2021 Mobvoi Inc. (authors: Di Wu)
#               2022 Tinnove Inc (authors: Wei Ren)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import argparse
import copy
import logging
import os
import sys

import torch
import yaml
from torch.utils.data import DataLoader
from textgrid import TextGrid, IntervalTier
import math

from wenet.dataset.dataset import Dataset
from wenet.utils.file_utils import read_symbol_table, read_non_lang_symbols
from wenet.utils.ctc_utils import force_align
from wenet.utils.common import get_subsample
from wenet.utils.init_model import init_model


def generator_textgrid(maxtime, lines, output):
    # Download Praat: https://www.fon.hum.uva.nl/praat/
    interval = maxtime / (len(lines) + 1)
    margin = 0.0001

    tg = TextGrid(maxTime=maxtime)
    linetier = IntervalTier(name="line", maxTime=maxtime)

    i = 0
    for l in lines:
        s, e, w = l.split()
        linetier.add(minTime=float(s) + margin, maxTime=float(e), mark=w)

    tg.append(linetier)
    print("successfully generator {}".format(output))
    tg.write(output)


def get_frames_timestamp(alignment,
                         prob,
                         blank_thres=0.999,
                         thres=0.0000000001):
    # convert alignment to a praat format, which is a doing phonetics
    # by computer and helps analyzing alignment
    timestamp = []
    # get frames level duration for each token
    start = 0
    end = 0
    local_start = 0
    while end < len(alignment):
        while end < len(alignment) and alignment[end] == 0:
            end += 1
        if end == len(alignment):
            timestamp[-1] += alignment[start:]
            break
        end += 1
        while end < len(alignment) and alignment[end - 1] == alignment[end]:
            end += 1
        local_start = end - 1
        # find the possible front border for current token
        while local_start >= start and (prob[local_start][0] < math.log(
                blank_thres) or prob[local_start][alignment[
                    end - 1]] > math.log(thres)):
            alignment[local_start] = alignment[end - 1]
            local_start -= 1
        cur_alignment = alignment[start:end]
        timestamp.append(cur_alignment)
        start = end
    return timestamp


def get_labformat(timestamp, subsample):
    begin = 0
    begin_time = 0
    duration = 0
    labformat = []
    for idx, t in enumerate(timestamp):
        # 25ms frame_length,10ms hop_length, 1/subsample
        subsample = get_subsample(configs)
        # time duration
        i = 0
        while t[i] == 0:
            i += 1
        begin = i
        dur = 0
        while i < len(t) and t[i] != 0:
            i += 1
            dur += 1
        begin = begin_time + begin * 0.01 * subsample
        duration = dur * 0.01 * subsample
        if idx < len(timestamp) - 1:
            print("{:.2f} {:.2f} {}".format(begin, begin + duration,
                                            char_dict[t[-1]]))
            labformat.append("{:.2f} {:.2f} {}\n".format(
                begin, begin + duration, char_dict[t[-1]]))
        else:  # last token
            non_blank = 0
            for i in t:
                if i != 0:
                    token = i
                    break
            print("{:.2f} {:.2f} {}".format(begin, begin + duration,
                                            char_dict[token]))
            labformat.append("{:.2f} {:.2f} {}\n".format(
                begin, begin + duration, char_dict[token]))
        begin_time += len(t) * 0.01 * subsample
    return labformat


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='use ctc to generate alignment')
    parser.add_argument('--config', required=True, help='config file')
    parser.add_argument('--input_file', required=True, help='format data file')
    parser.add_argument('--data_type',
                        default='raw',
                        choices=['raw', 'shard'],
                        help='train and cv data type')
    parser.add_argument('--gpu',
                        type=int,
                        default=-1,
                        help='gpu id for this rank, -1 for cpu')
    parser.add_argument('--blank_thres',
                        default=0.999999,
                        type=float,
                        help='ctc blank thes')
    parser.add_argument('--thres',
                        default=0.000001,
                        type=float,
                        help='ctc non blank thes')
    parser.add_argument('--checkpoint', required=True, help='checkpoint model')
    parser.add_argument('--dict', required=True, help='dict file')
    parser.add_argument(
        '--non_lang_syms',
        help="non-linguistic symbol file. One symbol per line.")
    parser.add_argument('--result_file',
                        required=True,
                        help='alignment result file')
    parser.add_argument('--batch_size', type=int, default=1, help='batch size')
    parser.add_argument('--gen_praat',
                        action='store_true',
                        help='convert alignment to a praat format')
    parser.add_argument('--bpe_model',
                        default=None,
                        type=str,
                        help='bpe model for english part')

    args = parser.parse_args()
    print(args)
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)s %(message)s')
    os.environ['CUDA_VISIBLE_DEVICES'] = str(args.gpu)

    if args.batch_size > 1:
        logging.fatal('alignment mode must be running with batch_size == 1')
        sys.exit(1)

    with open(args.config, 'r') as fin:
        configs = yaml.load(fin, Loader=yaml.FullLoader)

    # Load dict
    char_dict = {}
    with open(args.dict, 'r') as fin:
        for line in fin:
            arr = line.strip().split()
            assert len(arr) == 2
            char_dict[int(arr[1])] = arr[0]
    eos = len(char_dict) - 1

    symbol_table = read_symbol_table(args.dict)

    # Init dataset and data loader
    ali_conf = copy.deepcopy(configs['dataset_conf'])

    ali_conf['filter_conf']['max_length'] = 102400
    ali_conf['filter_conf']['min_length'] = 0
    ali_conf['filter_conf']['token_max_length'] = 102400
    ali_conf['filter_conf']['token_min_length'] = 0
    ali_conf['filter_conf']['max_output_input_ratio'] = 102400
    ali_conf['filter_conf']['min_output_input_ratio'] = 0
    ali_conf['speed_perturb'] = False
    ali_conf['spec_aug'] = False
    ali_conf['spec_trim'] = False
    ali_conf['shuffle'] = False
    ali_conf['sort'] = False
    ali_conf['fbank_conf']['dither'] = 0.0
    ali_conf['batch_conf']['batch_type'] = "static"
    ali_conf['batch_conf']['batch_size'] = args.batch_size
    non_lang_syms = read_non_lang_symbols(args.non_lang_syms)

    ali_dataset = Dataset(args.data_type,
                          args.input_file,
                          symbol_table,
                          ali_conf,
                          args.bpe_model,
                          non_lang_syms,
                          partition=False)

    ali_data_loader = DataLoader(ali_dataset, batch_size=None, num_workers=0)

    # Init asr model from configs
    model, configs = init_model(args, configs)

    use_cuda = args.gpu >= 0 and torch.cuda.is_available()
    device = torch.device('cuda' if use_cuda else 'cpu')
    model = model.to(device)

    model.eval()
    with torch.no_grad(), open(args.result_file, 'w',
                               encoding='utf-8') as fout:
        for batch_idx, batch in enumerate(ali_data_loader):
            print("#" * 80)
            key, feat, target, feats_length, target_length = batch

            feat = feat.to(device)
            target = target.to(device)
            feats_length = feats_length.to(device)
            target_length = target_length.to(device)
            # Let's assume B = batch_size and N = beam_size
            # 1. Encoder
            encoder_out, encoder_mask = model._forward_encoder(
                feat, feats_length)  # (B, maxlen, encoder_dim)
            maxlen = encoder_out.size(1)
            ctc_probs = model.ctc.log_softmax(
                encoder_out)  # (1, maxlen, vocab_size)
            # print(ctc_probs.size(1))
            ctc_probs = ctc_probs.squeeze(0)
            target = target.squeeze(0)
            alignment = force_align(ctc_probs, target)
            fout.write('{} {}\n'.format(key[0], alignment))

            if args.gen_praat:
                timestamp = get_frames_timestamp(alignment, ctc_probs,
                                                 args.blank_thres, args.thres)
                subsample = get_subsample(configs)
                labformat = get_labformat(timestamp, subsample)

                lab_path = os.path.join(os.path.dirname(args.result_file),
                                        key[0] + ".lab")
                with open(lab_path, 'w', encoding='utf-8') as f:
                    f.writelines(labformat)

                textgrid_path = os.path.join(os.path.dirname(args.result_file),
                                             key[0] + ".TextGrid")
                generator_textgrid(maxtime=(len(alignment) + 1) * 0.01 *
                                   subsample,
                                   lines=labformat,
                                   output=textgrid_path)



from gxl_ai_utils.utils import utils_data, utils_file
import os

input_dir = '/home/work_nfs6/disk2/zhyyao/data_16k/src/ench_haoweilai_587h'
output_dir = '/home/work_nfs7/xlgeng/data/asr_data_inventory/haoweilai_587h'

def make_total_text_wavscp():
    utils_file.makedir(output_dir)
    txt_scp_dict = utils_file.load_dict_from_scp(os.path.join(input_dir, 'label'))
    wav_scp_dict = utils_file.get_scp_for_wav_dir(os.path.join(input_dir, 'cs_wav'))
    utils_file.write_dict_to_scp(txt_scp_dict, os.path.join(output_dir, 'text'))
    utils_file.write_dict_to_scp(wav_scp_dict, os.path.join(output_dir, 'wav.scp'))

if __name__ == '__main__':
    make_total_text_wavscp()



import os

import torch
from matplotlib import pyplot as plt
from gxl_ai_utils import AiConstant

from . import store_model_class
from . import store_model_name


def load_model(name, **kwargs):
    return store_model_class.store_model_class.get_model_by_name(name, **kwargs)


def loader_checkpoint_by_modelname_epoch(model_name, epochs=None):
    model_dir = AiConstant.OUTPUT_PATH + model_name + "/"
    file_name_list = os.listdir(model_dir)
    file_name_list = [x for x in file_name_list if x.startswith(model_name + '_epoch')]
    if epochs:
        file_with_the_epoch = [x for x in file_name_list if x.startswith(model_name + '_epoch-' + str(epochs))]
        if len(file_with_the_epoch) != 0:
            model_path = model_dir + file_with_the_epoch[-1]
            return loader_checkpoint_by_path(model_path)
    if len(file_name_list) != 0:
        model_path = model_dir + file_name_list[-1]
        return loader_checkpoint_by_path(model_path)
    print('找不到模型参数文件,from store_model.py')
    return None


def loader_checkpoint_by_path(local=AiConstant.OUTPUT_PATH + 'mnist_0_epoch2.pth'):
    outpath = AiConstant.OUTPUT_PATH
    file_name = local.split('/').pop()
    parts = file_name.split('_')
    if len(parts) <= 1:
        print('文件名格式不正确,文件名应为 模型名_epoch{i}.pth,from store_model.py')
        return None
    else:
        parts.pop()
        model_name = '_'.join(parts)
        model = load_model(model_name)
        if model is None:
            print("未找到被加载的模型")
            return None
        # Load the model state_dict
        loaded_state_dict = torch.load(local, map_location=torch.device('cpu'))
        # Remove 'module.' prefix from keys
        new_state_dict = {k.replace('module.', ''): v for k, v in loaded_state_dict.items()}
        # Load the new state_dict into the model
        model.load_state_dict(new_state_dict)
        return model


if __name__ == '__main__':
    """"""
    # test_loader_model()



import sys
import time
import torch

from gxl_ai_utils import utils
from gxl_ai_utils.config.gxl_config import GxlNode
from gxl_ai_utils.run.base_train_runner import GxlBaseRunner
from gxl_ai_utils.AiConstant import AI_LOGGER
from gxl_ai_utils.utils import utils_model, utils_file


class RunnerGxl(GxlBaseRunner):
    def __init__(self, model, optim, loss_f, train_loader, config: GxlNode,
                 valid_loader=None, scheduler=None, multi=False,
                 local_rank=0, is_class=True,
                 device=torch.device('cpu')):
        super(RunnerGxl, self).__init__(model, optim, loss_f, train_loader,
                                        config, valid_loader, scheduler, multi,
                                        local_rank, is_class, device)

    def handle_batch(self, batch):
        X, y = batch
        X, y = utils_model.put_data_to_device(X, y, self.device)
        out = self.model(X)
        loss = self.loss_f(out, y)
        loss.backward()
        self.optim.step()
        if self.scheduler is not None:
            self.scheduler.step()
        return loss

    def calculate_valid_loss(self):
        accumulate_loss = 0.0
        with torch.no_grad:
            for i, X, y in enumerate(self.valid_loader):
                self.model.eval()
                X, y = utils_model.put_data_to_device(X, y, self.device)
                out = self.model(X)
                accumulate_loss += self.loss_f(out, y)
        return accumulate_loss / len(self.valid_loader)

    def calculate_valid_accuracy(self, dataloader: torch.utils.data.DataLoader = None):
        self.model.eval()
        all_accuracy = 0.0
        batch_num = 0
        with torch.no_grad():
            for X, y in self.valid_loader:
                X, y = utils_model.put_data_to_device(X, y, self.device)
                out = self.model(X)
                accuracy = self.accuracy(out, y)
                all_accuracy += accuracy
                batch_num += 1
        return all_accuracy / batch_num

    def inference(self, dataloader: torch.utils.data.DataLoader, file_save_path: str = './output/inference.jsonl'):
        self.model.eval()
        dic_list = []
        with torch.no_grad():
            for X, _, y in dataloader:
                X, y = utils_model.put_data_to_device(X, y, self.device)
                out = self.model(X)
                res = out.argmax(dim=1)
                assert len(res) == len(y)
                for re, y_i in zip(res, y):
                    res = {'true': y_i.item(), 'pred': re.item()}
                    dic_list.append(res)
        utils_file.write_dict_list_to_jsonl(dic_list, file_save_path)
        accuracy = self.calculate_valid_accuracy(dataloader)
        str = f"该次测试集的正确率为{accuracy * 100:.6f}%"
        with open(file_save_path, 'a') as f:
            f.write('\n')
            f.write(str)
        print(str)


if __name__ == '__main__':
    """"""
